"""
OpenMM system extensions and tools
==================================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains implementations of common OpenMM system
transformations, like the slab correction method for pseudo-2D slab
systems, the method of image charges, and an applied potential
difference.
"""

import logging
from typing import Any, Union
import warnings

import mpmath
import numpy as np
import openmm
from openmm import app, unit
from scipy import special

from .unit import VACUUM_PERMITTIVITY

try:
    from openmm_ic import ICLangevinIntegrator

    FOUND_ICPLUGIN = True
except ImportError:
    try:
        from constvplugin import ConstVLangevinIntegrator as ICLangevinIntegrator

        FOUND_ICPLUGIN = True
    except ImportError:
        FOUND_ICPLUGIN = False


def register_particles(
    system: openmm.System,
    topology: app.Topology,
    N: int = 0,
    mass: Union[float, unit.Quantity] = 0.0,
    *,
    chain: app.Chain = None,
    element: app.Element = None,
    name: str = "",
    resname: str = "",
    nbforce: openmm.NonbondedForce = None,
    charge: Union[float, unit.Quantity] = 0.0,
    sigma: Union[float, unit.Quantity] = 0.0,
    epsilon: Union[float, unit.Quantity] = 0.0,
    cnbforces: dict[openmm.CustomNonbondedForce, tuple[Any]] = None,
) -> None:
    """
    Sequentially registers particles of the same type to the simulation
    system, topology, and nonbonded pair potentials.

    Parameters
    ----------
    system : `openmm.System`
        OpenMM molecular system. Can be :code:`None` if particles are to
        be registered to the topology only.

    topology : `openmm.app.Topology`
        Topological information about an OpenMM system.

    N : `int`, default: :code:`0`
        Number of atom(s). If not specified, no particles are added.

    mass : `float` or `openmm.unit.Quantity`, default: :code:`0`
        Molar mass. If not specified, particles are massless and will
        not move.

        **Reference unit**: :math:`\\mathrm{g/mol}`.

    chain : `openmm.app.Chain`, keyword-only, optional
        Chain that the atom(s) should be added to. If not provided,
        a new chain is created for each atom.

    element : `openmm.app.Element`, keyword-only, optional
        Chemical element of the atom(s) to add.

    name : `str`, keyword-only, optional
        Name of the atom(s) to add.

    resname : `str`, keyword-only, optional
        Name of the residue(s) to add. If not specified, `name` is used
        if available.

    nbforce : `openmm.NonbondedForce`, keyword-only, optional
        Standard OpenMM pair potential object implementing the nonbonded
        Lennard-Jones and Coulomb potentials.

    charge : `float` or `openmm.unit.Quantity`, keyword-only, \
    default: :code`0`
        Charge :math:`q` of the atom(s) for use in the Coulomb potential
        in `nbforce`.

        **Reference unit**: :math:`\\mathrm{e}`.

    sigma : `float` or `openmm.unit.Quantity`, keyword-only, \
    default: :code`0`
        :math:`\\sigma` parameter of the Lennard-Jones potential in
        `nbforce`.

        **Reference unit**: :math:`\\mathrm{nm}`.

    epsilon : `float` or `openmm.unit.Quantity`, keyword-only, \
    default: :code`0`
        :math:`\\epsilon` parameter of the Lennard-Jones potential in
        `nbforce`.

        **Reference unit**: :math:`\\mathrm{kJ/mol}`.

    cnbforces : `dict`, keyword-only, optional
        Custom pair potential objects implementing other non-standard
        pair potentials and their corresponding per-particle parameters.

        **Example**: :code:`{gauss: (0.3 * unit.nanometer,)}`, where
        `gauss` is a custom Gaussian potential obtained using
        :func:`mdcraft.openmm.pair.gauss()`.
    """

    has_nbforce = nbforce is not None
    has_system = system is not None
    per_chain = chain is None
    cnbforces = cnbforces or {}
    for _ in range(N):
        if has_system:
            system.addParticle(mass)
        if per_chain:
            chain = topology.addChain()
        residue = topology.addResidue(resname or name, chain)
        topology.addAtom(name, element, residue)
        if has_nbforce:
            nbforce.addParticle(charge, sigma, epsilon)
        for cnbforce, param in cnbforces.items():
            cnbforce.addParticle(param)


def add_slab_correction(
    system: openmm.System,
    topology: app.Topology,
    nbforce: Union[openmm.NonbondedForce, openmm.CustomNonbondedForce],
    temp: Union[float, unit.Quantity],
    fric: Union[float, unit.Quantity],
    dt: Union[float, unit.Quantity],
    axis: int = 2,
    *,
    charge_index: int = 0,
    z_scale: float = 3,
    method: str = "force",
) -> openmm.Integrator:
    r"""
    Implements a slab correction so that efficient three-dimensional
    Ewald methods can continue to be used to evaluate the electrostatics
    for systems that are periodic in the :math:`x`- and
    :math:`y`-directions but not the :math:`z`-direction. Effectively,
    the system is treated as if it were periodic in the
    :math:`z`-direction, but with empty volume added between the slabs
    and the slab–slab interactions removed.

    For electroneutral systems, the Yeh–Berkowitz correction [1]_ is
    applied:

    .. math::

       \begin{gather*}
         U^\mathrm{corr}=\frac{N_A}{2\varepsilon_0V}M_z^2\\
         u_i^\mathrm{corr}=\frac{N_A}{2\varepsilon_0V}q_i\left(z_iM_z
         -\frac{\sum_i q_iz_i^2}{2}\right)\\
         f_{i,z}^\mathrm{corr}=-\frac{N_A}{\varepsilon_0V}q_iM_z
       \end{gather*}

    For systems with a net electric charge, the Ballenegger–Arnold–Cerdà
    correction [2]_ is applied instead:

    .. math::

       \begin{gather*}
         U^\mathrm{corr}=\frac{N_A}{2\varepsilon_0V}
         \left(M_z^2-q_\mathrm{tot}\sum_i q_iz_i^2
         -\frac{q_\mathrm{tot}^2L_z^2}{12}\right)\\
         u_i^\mathrm{corr}=\frac{N_A}{2\varepsilon_0V}q_i
         \left(z_iM_z-\frac{\sum_i q_iz_i^2+q_\mathrm{tot}z_i^2}{2}
         -\frac{q_\mathrm{tot}L_z^2}{12}\right)\\
         f_{i,z}^\mathrm{corr}=-\frac{N_A}{\varepsilon_0V}q_i
         \left(M_z-q_\mathrm{tot}z_i\right)
       \end{gather*}

    Note that the the relative permittivity
    :math:`\varepsilon_\mathrm{r}` does not appear in the equations
    above because it is accounted for by scaling the particle charges in
    the Coulomb potential.

    Parameters
    ----------
    system : `openmm.System`
        OpenMM molecular system.

    topology : `openmm.app.Topology`
        Topological information about an OpenMM system.

    nbforce : `openmm.NonbondedForce` or `openmm.CustomNonbondedForce`
        Pair potential object containing particle charge information.

        .. note::

           It is assumed that the charge :math:`q` information is the
           first per-particle parameter stored in `nbforce`. If not, the
           index can be specified in `charge_index`.

    temp : `float` or `openmm.unit.Quantity`
        System temperature :math:`T`.

        **Reference unit**: :math:`\mathrm{K}`.

    fric : `float` or `openmm.unit.Quantity`
        Friction coefficient :math:`\gamma` that couples the system to
        the heat bath.

        **Reference unit**: :math:`\mathrm{ps}^{-1}`.

    dt : `float` or `openmm.unit.Quantity`
        Integration step size :math:`\Delta t`.

        **Reference unit**: :math:`\mathrm{ps}`.

    axis : `int`, default: :code:`2`
        Axis along which to apply the slab correction, with :math:`x`
        being :code:`0`, :math:`y` being :code:`1`, and :math:`z` being
        :code:`2`. The source code and outputs, if any, will refer to
        this dimension as :code:`z` regardless of this value.

    charge_index : `int`, keyword-only, default: :code:`0`
        Index of charge :math:`q` information in the per-particle
        parameters stored in `nbforce`.

    z_scale : `float`, keyword-only, default: :code:`3`
        Scaling factor for the dimension specified in `axis`.

    method : `str`, keyword-only, default: :code:`"force"`
        Slab correction methodology.

        .. container::

           **Valid values**:

           * :code:`"force"`: Collective implementation via
             :class:`openmm.CustomExternalForce` and
             :class:`openmm.CustomCVForce` [3]_. This is
             generally the most efficient.
           * :code:`"integrator"`: Per-particle implementation via an
             :class:`openmm.CustomIntegrator`.

    Returns
    -------
    integrator : `openmm.Integrator` or `openmm.CustomIntegrator`
        Integrator that simulates a system using Langevin dynamics, with
        the LFMiddle discretization.

    References
    ----------
    .. [1] Yeh, I.-C.; Berkowitz, M. L. Ewald Summation for Systems with
       Slab Geometry. *J. Chem. Phys.* **1999**, *111* (7), 3155–3162.
       https://doi.org/10.1063/1.479595.

    .. [2] Ballenegger, V.; Arnold, A.; Cerdà, J. J. Simulations of
       Non-Neutral Slab Systems with Long-Range Electrostatic
       Interactions in Two-Dimensional Periodic Boundary Conditions.
       *J. Chem. Phys.* **2009**, *131* (9), 094107.
       https://doi.org/10.1063/1.3216473.

    .. [3] Son, C. Y.; Wang, Z.-G. Image-Charge Effects on Ion
       Adsorption near Aqueous Interfaces. *Proc. Natl. Acad. Sci.
       U.S.A.* **2021**, *118* (19), e2020615118.
       https://doi.org/10.1073/pnas.2020615118.
    """

    # Get system dimensions
    dims = (
        np.array(topology.getUnitCellDimensions().value_in_unit(unit.nanometer))
        * unit.nanometer
    )
    pbv = system.getDefaultPeriodicBoxVectors()

    # Scale system z-dimension by specified z-scale
    if z_scale < 2:
        wmsg = (
            "A z-scaling factor that is less than 2 may introduce "
            "unwanted slab–slab interactions. The recommended "
            "value is 3."
        )
        warnings.warn(wmsg)
    elif z_scale > 5:
        wmsg = (
            "A z-scaling factor that is greater than 5 may "
            "penalize performance. The recommended value is 3."
        )
        warnings.warn(wmsg)
    dims[axis] *= z_scale
    pbv[axis] *= z_scale

    # Set new system dimensions
    topology.setUnitCellDimensions(dims)
    system.setDefaultPeriodicBoxVectors(*pbv)

    # Obtain particle charge information
    if isinstance(nbforce.getParticleParameters(0)[charge_index], unit.Quantity):
        qs = np.fromiter(
            (
                nbforce.getParticleParameters(i)[charge_index].value_in_unit(
                    unit.elementary_charge
                )
                for i in range(nbforce.getNumParticles())
            ),
            dtype=float,
        )
    else:
        qs = np.fromiter(
            (
                nbforce.getParticleParameters(i)[charge_index]
                for i in range(nbforce.getNumParticles())
            ),
            dtype=float,
        )
    neutral = qs.min() == qs.max()
    if not neutral:
        q_tot = qs.sum()
        electroneutral = np.isclose(q_tot, 0)

    # Calculate coefficient for slab correction
    coef = unit.AVOGADRO_CONSTANT_NA / (
        2 * VACUUM_PERMITTIVITY * dims[0] * dims[1] * dims[2]
    )

    # Get letter representation of axis for formula
    z = chr(120 + axis)

    if neutral:

        # Instantiate an integrator that simulates a system using
        # Langevin dynamics, with the LFMiddle discretization
        integrator = openmm.LangevinMiddleIntegrator(temp, fric, dt)

    else:
        if method == "integrator":

            # Implement an integrator that simulates a system using Langevin
            # dynamics, with the LFMiddle discretization
            integrator = openmm.CustomIntegrator(dt)
            integrator.addGlobalVariable("a", np.exp(-fric * dt))
            integrator.addGlobalVariable("b", np.sqrt(1 - np.exp(-2 * fric * dt)))
            integrator.addGlobalVariable(
                "kT", unit.AVOGADRO_CONSTANT_NA * unit.BOLTZMANN_CONSTANT_kB * temp
            )
            integrator.addPerDofVariable("x1", 0)
            integrator.addUpdateContextState()
            integrator.addComputePerDof("v", "v+dt*f/m")
            integrator.addConstrainVelocities()
            integrator.addComputePerDof("x", "x+dt*v/2")
            integrator.addComputePerDof("v", "a*v+b*sqrt(kT/m)*gaussian")
            integrator.addComputePerDof("x", "x+dt*v/2")
            integrator.addComputePerDof("x1", "x")
            integrator.addConstrainPositions()
            integrator.addComputePerDof("v", "v+(x-x1)/dt")

            # Initialize per-degree-of-freedom variable q for charge
            integrator.addPerDofVariable("q", 0)

            # Add global dipole moment computation to integrator
            integrator.addComputeSum("M_z", "q*x")
            integrator.addComputeSum("M_zz", "q*x^2")

            # Give particle charge information to integrator
            q_vectors = np.zeros((len(qs), 3))
            q_vectors[:, axis] = qs
            integrator.setPerDofVariableByName("q", q_vectors)

            # Implement per-particle slab correction
            if electroneutral:
                slab_corr = openmm.CustomExternalForce(f"coef*q*({z}*M_z-M_zz/2)")
            else:
                slab_corr = openmm.CustomExternalForce(
                    f"coef*q*({z}*M_z-(M_zz+q_tot*{z}^2)/2-q_tot*dim_z^2/12)"
                )
                slab_corr.addGlobalParameter("dim_z", dims[axis])
                slab_corr.addGlobalParameter("q_tot", q_tot)
            slab_corr.addGlobalParameter("M_z", 0)
            slab_corr.addGlobalParameter("M_zz", 0)
            slab_corr.addGlobalParameter("coef", coef)
            slab_corr.addPerParticleParameter("q")

            # Register real particles to the slab correction
            for i, q in enumerate(qs):
                slab_corr.addParticle(i, (q,))

        elif method == "force":

            # Instantiate an integrator that simulates a system using
            # Langevin dynamics, with the LFMiddle discretization
            integrator = openmm.LangevinMiddleIntegrator(temp, fric, dt)

            # Calculate instantaneous system dipole
            M_z = openmm.CustomExternalForce(f"q*{z}")
            M_z.addPerParticleParameter("q")

            # Implement collective slab correction
            if electroneutral:
                slab_corr = openmm.CustomCVForce("coef*M_z^2")
            else:
                M_zz = openmm.CustomExternalForce(f"q*{z}^2")
                M_zz.addPerParticleParameter("q")
                slab_corr = openmm.CustomCVForce(
                    "coef*(M_z^2-q_tot*M_zz-q_tot^2*dim_z^2/12)"
                )
                slab_corr.addCollectiveVariable("M_zz", M_zz)
                slab_corr.addGlobalParameter("dim_z", dims[axis])
                slab_corr.addGlobalParameter("q_tot", q_tot)
            slab_corr.addCollectiveVariable("M_z", M_z)
            slab_corr.addGlobalParameter("coef", coef)

            # Register real particles to the slab correction
            for i, q in enumerate(qs):
                M_z.addParticle(i, (q,))
                if not electroneutral:
                    M_zz.addParticle(i, (q,))

        # Register slab correction to the system
        system.addForce(slab_corr)
        logging.info("Slab correction applied to the system.")

    return integrator


def add_image_charges(
    system: openmm.System,
    topology: app.Topology,
    positions: Union[np.ndarray[float], unit.Quantity],
    temp: Union[float, unit.Quantity],
    fric: Union[float, unit.Quantity],
    dt: Union[float, unit.Quantity],
    *,
    gamma: float = -1,
    n_cells: int = 2,
    nbforce: openmm.NonbondedForce = None,
    cnbforces: dict[openmm.CustomNonbondedForce, dict[str, Any]] = None,
    wall_indices: np.ndarray[int] = None,
    exclude: bool = False,
) -> tuple[unit.Quantity, openmm.Integrator]:
    r"""
    Implements the method of image charges for dielectric boundaries.
    For more information about the method, see Refs. [1]_, [2]_, and
    [3]_ for perfectly conducting boundaries, and Refs. [4]_, [5]_, and
    [6]_ otherwise.

    .. note::

       The boundaries must be in the :math:`xy`-plane and along the
       :math:`z`-axis.

    Parameters
    ----------
    system : `openmm.System`
        OpenMM molecular system.

    topology : `openmm.app.Topology`
        Topological information about an OpenMM system.

    positions : `numpy.ndarray`
        Positions of the :math:`N` particles in the real system.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    temp : `float` or `openmm.unit.Quantity`
        System temperature :math:`T`.

        **Reference unit**: :math:`\mathrm{K}`.

    fric : `float` or `openmm.unit.Quantity`
        Friction coefficient :math:`\gamma` that couples the system to
        the heat bath.

        **Reference unit**: :math:`\mathrm{ps}^{-1}`.

    dt : `float` or `openmm.unit.Quantity`
        Integration step size :math:`\Delta t`.

        **Reference unit**: :math:`\mathrm{ps}`.

    gamma : `float`, keyword-only, default: :code:`-1`
        Scaled image charge magnitude :math:`\gamma`.

    n_cells : `int`, keyword-only, default: :code:`2`
        Total number :math:`n_\mathrm{cell}` of cells, with :math:`1`
        real and :math:`n_\mathrm{cell}-1` image cells.

        .. tip::

           Use :math:`n_\mathrm{cell}>2` to instantiate a
           highly-confined simulation system with periodic dimensions
           large enough to satisfy the constraint that the force field
           cutoff be less than half the box size.

        **Valid values**: Even integers greater than :math:`1` when
        :math:`\gamma=-1`.

    nbforce : `openmm.NonbondedForce`, keyword-only, optional
        Standard OpenMM pair potential object implementing the nonbonded
        Lennard-Jones and Coulomb potentials. For the image charges, the
        charges :math:`q` have their signs flipped, and the LJ
        parameters :math:`\sigma` and :math:`\epsilon` are both set to
        :math:`0`.

    cnbforces : `dict`, keyword-only, optional
        Custom pair potential objects implementing other non-standard
        pair potentials. The keys are the potentials and the values are
        the corresponding per-particle parameters to use, which can be
        provided as follows:

        .. container::

           * `None`: All per-particle parameters set to :math:`0`.
           * `dict`: Information on how to transform the real particle
             parameters into those for the image charges. The valid
             (optional) key-value pairs are:

             * :code:`"charge"`: Index (`int`) where the charge
               :math:`q` information is stored.
             * :code:`"zero"`: Index (`int`) or indices (`list`,
               `slice`, or `numpy.ndarray`) of parameters to zero out.
             * :code:`"replace"`: `dict` containing key-value pairs of
               indices (`int`) of parameters to change and their
               replacement value(s). If the value is an `int`, all
               particles receive that value for the current pair
               potential. If the value is another `dict`, the new
               parameter value (value of inner `dict`) is determined by
               the current parameter value (key of inner `dict`).

        To prevent unintended behavior, ensure that each parameter index
        for a custom pair potential is used only once across the keys
        listed above. If one does appear more than once, the order
        in which parameter values are updated follows that of the list
        above.

        See the Examples section below for a simple illustration of the
        possibilities outlined above.

    wall_indices : array-like, keyword-only, optional
        Atom indices corresponding to wall particles. If not provided,
        the wall particles are guessed using the system dimensions.

    exclude : `bool`, keyword-only, default: :code:`False`
        Specifies whether interactions between a wall particle and all
        image wall particles should be excluded. If :code:`False`, only
        the interaction between a wall particle and its own image is
        removed.

        .. tip::

           If you want accurate forces acting on wall particles, set
           :code:`exclude=True`.

    Returns
    -------
    positions : `numpy.ndarray`
        Positions of the :math:`N` particles in the real system and
        their images.

        **Shape**: :math:`(2N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    integrator : `openmm.Integrator`
        Integrator that simulates a system using Langevin dynamics, with
        the LFMiddle discretization.

    Examples
    --------
    Prepare the `cnbforces` `dict` argument for a system with the
    following custom pair potentials:

    * :code:`pair_gauss`: Gaussian potential with per-particle
      parameters `type` and `beta`.

      * `type` is the particle type, and is used to look up a
        tabulated function for parameter values. As such, the new
        values will vary and depend on the old ones. For example, for
        a system with :math:`2` types of particles, :math:`0` becomes
        :math:`2` and :math:`1` becomes :math:`3`.
      * `beta` is assumed to be constant, with the value for the real
        particles differing from that for image charges. For this
        example, the new value is :math:`\beta = 42` (arbitrary
        units).

    * :code:`pair_wca`: Weeks–Chander–Andersen potential with per-
      particle parameters `sigma` and `epsilon`, both of which
      should be set to :math:`0` to turn off this potential for
      image charges.
    * :code:`pair_coul_recip`: Smeared Coulomb potential with per-
      particle parameter `charge`, which should be flipped for image
      charges, and `dummy`, a value that should be set to :math:`0`.

    .. code::

       NEW_BETA_VALUE = 42
       cnbforces = {
           pair_gauss: {"replace": {0: {0: 2, 1: 3}, 1: NEW_BETA_VALUE}},
           pair_wca: None,
           pair_coul_recip: {"charge": 0, "zero": 1}
       }

    References
    ----------
    .. [1] Hautman, J.; Halley, J. W.; Rhee, Y. ‐J. Molecular Dynamics
       Simulation of Water Beween Two Ideal Classical Metal Walls.
       *J. Chem. Phys.* **1989**, *91* (1), 467–472.
       https://doi.org/10.1063/1.457481.

    .. [2] Dwelle, K. A.; Willard, A. P. Constant Potential,
       Electrochemically Active Boundary Conditions for Electrochemical
       Simulation. *J. Phys. Chem. C* **2019**, *123* (39), 24095–24103.
       https://doi.org/10.1021/acs.jpcc.9b06635.

    .. [3] Son, C. Y.; Wang, Z.-G. Image-Charge Effects on Ion
       Adsorption near Aqueous Interfaces. *Proc. Natl. Acad. Sci.
       U.S.A.* **2021**, *118* (19), e2020615118.
       https://doi.org/10.1073/pnas.2020615118.

    .. [4] Dos Santos, A. P.; Girotto, M.; Levin, Y. Simulations of
       Coulomb Systems with Slab Geometry Using an Efficient 3D Ewald
       Summation Method. *The Journal of Chemical Physics* **2016**,
       **144** (14), 144103. https://doi.org/10.1063/1.4945560.

    .. [5] Dos Santos, A. P.; Girotto, M.; Levin, Y. Simulations of
       Coulomb Systems Confined by Polarizable Surfaces Using Periodic
       Green Functions. *The Journal of Chemical Physics* **2017**,
       **147** (18), 184105. https://doi.org/10.1063/1.4997420.

    .. [6] Son, C. Y.; Wang, Z.-G. Manuscript in preparation.
    """

    if not FOUND_ICPLUGIN:
        emsg = (
            "An integrator capable of simulating a system with "
            "image charges was not found. As such, the method of "
            "image charges is unavailable unless the openmm-ic "
            "(https://github.com/bbye98/mdcraft/tree/main/lib/openmm-ic-plugin) "
            "or the constvplugin (https://github.com/scychon/openmm_constV) "
            "package is installed."
        )
        raise ImportError(emsg)

    if np.isclose(gamma, 0):
        emsg = (
            "Use the slab correction, available via "
            "mdcraft.openmm.system.slab_correction(), for gamma=0."
        )
        raise ValueError(emsg)
    if not np.isclose(gamma, -1) and n_cells != 2:
        emsg = (
            "The method of image charges with gamma != -1 is only "
            "implemented for n_cells=2."
        )
        raise ValueError(emsg)

    def _ic_beta(gamma: float, x: float) -> float:
        r"""
        Computes the :math:`\beta` value used in the higher-order term
        correction for the method of image charges with
        :math:`\gamma\neq\pm1`.

        Parameters
        ----------
        gamma : `float`
            Scaled image charge magnitude :math:`\gamma`.

        x : `float`
            Scaled :math:`x`-coordinate of the ion.

        Returns
        -------
        beta : `float`
            Approximated :math:`\beta` value.
        """

        if not 0 <= x <= 1:
            raise ValueError("'x' must be between 0 and 1.")
        if np.isclose(x, 0.5):
            return float(
                2 * special.zeta(3, 1.5)
                - 2 * gamma**4 * mpmath.lerchphi(gamma**2, 3, 1.5)
            )
        else:
            return (
                special.zeta(2, 2 - x)
                - special.zeta(2, 1 + x)
                - gamma**4
                * float(
                    mpmath.lerchphi(gamma**2, 2, 2 - x)
                    - mpmath.lerchphi(gamma**2, 2, 1 + x)
                )
            ) / (2 * x - 1)

    # Get system information
    dims = (
        np.asarray(topology.getUnitCellDimensions().value_in_unit(unit.nanometer))
        * unit.nanometer
    )
    pbv = system.getDefaultPeriodicBoxVectors()
    N_real_atoms = positions.shape[0]
    if isinstance(positions, unit.Quantity):
        positions = positions.value_in_unit(unit.nanometer)

    # Guess indices of left and right walls if not provided
    if wall_indices is None:
        wall_indices = np.concatenate(
            (
                np.isclose(positions[:, 2], 0).nonzero()[0],
                np.isclose(
                    positions[:, 2], dims[2].value_in_unit(unit.nanometer)
                ).nonzero()[0],
            )
        )

    # Find averaged beta value for image charges with gamma =/= +/-1
    beta = (_ic_beta(gamma, 0) + _ic_beta(gamma, 0.5)) / 2

    # Set up higher-order image charge and slab corrections
    cv_E_corr = openmm.CustomExternalForce("q*(1-2*z/L)")
    cv_E_corr.addGlobalParameter("L", dims[2])  # real system z-dimension such
    # that 0 <= (x = z / L_z) <= 1
    cv_E_corr.addPerParticleParameter("q")
    cv_M_z = openmm.CustomExternalForce("q*z")
    cv_M_z.addPerParticleParameter("q")
    cv_M_zz = openmm.CustomExternalForce("q*z^2")
    cv_M_zz.addPerParticleParameter("q")

    # Obtain particle charge information and register charges to
    # corrections
    if nbforce is None:
        charge_index = None
        for force, params in cnbforces:
            if "charge" in params:
                charge_index = params["charge"]
                break
        if charge_index is None:
            raise ValueError("No charge information provided.")
    else:
        force = nbforce
        charge_index = 0

    q_tot = 0
    if isinstance(force.getParticleParameters(0)[charge_index], unit.Quantity):
        for i in range(force.getNumParticles()):
            q = force.getParticleParameters(i)[charge_index].value_in_unit(
                unit.elementary_charge
            )
            q_tot += q
            if not np.isclose(q, 0):
                cv_E_corr.addParticle(i, (q,))
                cv_M_z.addParticle(i, (q,))
                cv_M_zz.addParticle(i, (q,))
    else:
        for i in range(force.getNumParticles()):
            q = force.getParticleParameters(i)[charge_index]
            q_tot += q
            if not np.isclose(q, 0):
                cv_E_corr.addParticle(i, (q,))
                cv_M_z.addParticle(i, (q,))
                cv_M_zz.addParticle(i, (q,))
    electroneutral = np.isclose(q_tot, 0)

    # Update and set new system dimensions
    dims[2] *= n_cells
    topology.setUnitCellDimensions(dims)
    pbv[2] *= n_cells
    system.setDefaultPeriodicBoxVectors(*pbv)
    logging.info(f"Increased z-dimension to {dims[2]}.")

    # Determine correction energy expression
    corr_energy = ""
    corr = openmm.CustomCVForce(corr_energy)
    if not np.isclose(beta, 0):
        corr_energy += "coef1*E_corr*M_z"
        corr.addCollectiveVariable("E_corr", cv_E_corr)
        corr.addGlobalParameter(
            "coef1",
            (
                unit.AVOGADRO_CONSTANT_NA
                * gamma
                * beta
                / (4 * np.pi * VACUUM_PERMITTIVITY * dims[2] ** 2)
            ).in_units_of(
                unit.kilojoule_per_mole / (unit.elementary_charge**2 * unit.nanometer)
            ),
        )
    if not np.isclose(gamma, -1):
        corr_energy += "+coef2*M_z^2"
    if not electroneutral:
        if np.isclose(gamma, 1):
            corr_energy += "-coef2*q_tot*M_z*L_z"
        elif np.isclose(gamma, -1):
            corr_energy += "+coef2*q_tot*(M_z*L_z-M_zz)"
        else:
            corr_energy += "-coef2*q_tot*M_zz"
        corr.addGlobalParameter("q_tot", q_tot)
    if "coef2" in corr_energy:
        corr.addGlobalParameter(
            "coef2",
            (
                unit.AVOGADRO_CONSTANT_NA
                / (2 * VACUUM_PERMITTIVITY * dims[0] * dims[1] * dims[2])
            ).in_units_of(
                unit.kilojoule_per_mole / (unit.elementary_charge * unit.nanometer) ** 2
            ),
        )
    if "L_z" in corr_energy:
        corr.addGlobalParameter("L_z", dims[2])  # periodic system z-dimension
    if "M_z" in corr_energy:
        corr.addCollectiveVariable("M_z", cv_M_z)
    if "M_zz" in corr_energy:
        corr.addCollectiveVariable("M_zz", cv_M_zz)
    if corr_energy:
        if corr_energy.startswith("+"):
            corr_energy = corr_energy.lstrip("+")
        corr.setEnergyFunction(corr_energy)
        system.addForce(corr)
        logging.info("Added higher-order image charge and/or slab " "correction(s).")

    # Mirror particle positions
    if n_cells == 2:
        positions = (
            np.concatenate((positions, positions * np.array((1, 1, -1), dtype=int)))
            * unit.nanometer
        )
    else:
        positions = np.tile(positions, (n_cells, 1))
        for cell_index in range(1, n_cells):
            start = cell_index * N_real_atoms
            stop = (cell_index + 1) * N_real_atoms
            positions[start:stop, 2] = (1 - 2 * (cell_index % 2)) * positions[
                start:stop, 2
            ] - 2 * np.floor(cell_index / 2) * dims[2].value_in_unit(unit.nanometer)
        positions *= unit.nanometer
    logging.info(
        f"Replicated {N_real_atoms:,} particles {n_cells - 1} "
        "time(s) over the z-axis."
    )

    # Instantiate an integrator that simulates a system using
    # Langevin dynamics and updates the image charge positions
    integrator = ICLangevinIntegrator(temp, fric, dt, n_cells)

    # Register image charges to the system, topology, and force field
    cnbforces = cnbforces or {}
    N_real_chains = topology.getNumChains()
    atoms = list(topology.atoms())
    residues = list(topology.residues())
    coefs = (1, gamma)
    for c in range(1, n_cells):
        coef = coefs[c % 2]
        chains_ic = [topology.addChain() for _ in range(N_real_chains)]
        residues_ic = [
            topology.addResidue(f"IC_{r.name}", chains_ic[r.chain.index])
            for r in residues
        ]
        for i, atom in enumerate(atoms):
            system.addParticle(0)
            topology.addAtom(
                f"IC_{atom.name}", atom.element, residues_ic[atom.residue.index]
            )
            if nbforce is not None:
                nbforce.addParticle(
                    (
                        0
                        if i in wall_indices
                        else coef * nbforce.getParticleParameters(i)[0]
                    ),
                    0,
                    0,
                )
            for force, kwargs in cnbforces.items():
                params = np.array(force.getParticleParameters(i))
                if kwargs is None:
                    params[:] = 0
                else:
                    if "charge" in kwargs:
                        params[kwargs["charge"]] *= 0 if i in wall_indices else coef
                    if "zero" in kwargs:
                        params[kwargs["zero"]] = 0
                    if "replace" in kwargs:
                        for index, value in kwargs["replace"].items():
                            params[index] = (
                                value[params[index]]
                                if isinstance(value, dict)
                                else value
                            )
                force.addParticle(params)
    logging.info(
        f"Registered {system.getNumParticles() - N_real_atoms:,} "
        "image particles to the force field."
    )

    # Add existing particle exclusions to mirrored image charges
    for i in range(nbforce.getNumExceptions()):
        i1, i2, qq = nbforce.getExceptionParameters(i)[:3]
        if i1 not in wall_indices and i2 not in wall_indices:
            for c in range(1, n_cells):
                nbforce.addException(
                    c * N_real_atoms + i1, c * N_real_atoms + i2, qq, 0, 0
                )
                for force in cnbforces:
                    i1, i2 = force.getExclusionParticles(i)
                    force.addExclusion(c * N_real_atoms + i1, c * N_real_atoms + i2)
    logging.info(
        "Mirrored excluded non-wall image particle–image " "particle interactions."
    )

    # Prevent wall particles from interacting with all image wall
    # particles
    if exclude:
        for i in wall_indices:
            for j in wall_indices:
                for c in range(1, n_cells):
                    nbforce.addException(i, c * N_real_atoms + j, 0, 0, 0)
                    for force in cnbforces:
                        force.addExclusion(i, c * N_real_atoms + j)

    # Prevent wall particles from interacting their images
    else:
        for i in wall_indices:
            for c in range(1, n_cells):
                nbforce.addException(i, c * N_real_atoms + i, 0, 0, 0)
                for force in cnbforces:
                    force.addExclusion(i, c * N_real_atoms + i)
    logging.info("Removed wall–image wall interactions.")

    return positions, integrator


def add_electric_field(
    system: openmm.System,
    nbforce: openmm.NonbondedForce,
    E: Union[float, unit.Quantity],
    *,
    axis: int = 2,
    dielectric: float = 1,
    charge_index: int = 0,
    atom_indices: Union[int, np.ndarray[int]] = None,
) -> None:
    r"""
    Adds an electric field to all charged particles by adding a force
    :math:`f_i=q_iE` in the axis specified in `axis`, where :math:`q_i`
    is the per-particle charge and :math:`E` is the electric field.

    .. hint::

       The following schematic shows how directionality is handled:

       .. code::

          |-| (-) ---> |+|
          |-| <-- E -- |+|
          |-| <--- (+) |+|

       With a positive potential difference
       (:math:`\Delta V>0\;\mathrm{V}`), the electric field is negative
       (:math:`E<0\;\mathrm{V/m}`) such that it is pointing from the
       right (positive) plate to the left (negative) plate. If an ion
       has a positive charge (:math:`q_i>0\;\mathrm{e}`), the force will
       be negative, indicating that it will be drawn to the left plate,
       and vice versa.

    Parameters
    ----------
    system : `openmm.System`
        OpenMM molecular system.

    nbforce : `openmm.NonbondedForce` or `openmm.CustomNonbondedForce`
        Pair potential object containing particle charge information.

        .. note::

           It is assumed that the charge :math:`q` information is the
           first per-particle parameter stored in `nbforce`. If not, the
           index can be specified in `charge_index`.

    E : `float` or `openmm.unit.Quantity`
        Electric field :math:`E`.

        **Reference unit**: :math:`\mathrm{kJ/(mol\cdot nm\cdot e)}`.

    axis : `int`, keyword-only, default: :code:`2`
        Axis along which the walls are placed. :code:`0`, :code:`1`, and
        :code:`2` correspond to :math:`x`, :math:`y`, and :math:`z`,
        respectively.

    dielectric : `float`, keyword-only, default: :code:`1`
        Relative permittivity :math:`\varepsilon_\mathrm{r}` of the
        medium. Used to scale the particle charges by
        :math:`\sqrt{\varepsilon_\mathrm{r}}` and recover the original
        values.

    charge_index : `int`, keyword-only, default: :code:`0`
        Index of charge :math:`q` information in the per-particle
        parameters stored in `nbforce`.

    atom_indices : `int` or array-like, keyword-only, optional
        Indices of atoms to apply the electric field to. By default,
        the electric field is applied to all atoms, but this can be
        computationally expensive when there are charged particles that
        do not need to be included, such as image charges. If an `int`
        is provided, all atoms up to that index are included.
    """

    # Get letter representation of axis for formula
    z = chr(120 + axis)

    # Get indices of atoms that are affected by the electric field
    if atom_indices is None:
        atom_indices = range(nbforce.getNumParticles())
    elif isinstance(atom_indices, int):
        atom_indices = range(atom_indices)

    # Create and register particles to the electric field
    efield = openmm.CustomExternalForce(f"-q*E*{z}")
    efield.addGlobalParameter("E", E)
    efield.addPerParticleParameter("q")

    for i in atom_indices:
        q = nbforce.getParticleParameters(i)[charge_index]
        if isinstance(q, unit.Quantity):
            q = q.value_in_unit(unit.elementary_charge)
        if not np.isclose(q, 0):
            efield.addParticle(i, (q * np.sqrt(dielectric),))

    system.addForce(efield)


def estimate_pressure_tensor(
    context: openmm.Context, dh: float = 1e-5, *, diag: bool = False
) -> np.ndarray[float]:
    r"""
    Computes the estimated pressure tensor using a central finite
    difference for the virial contribution.

    The pressure tensor is given by

    .. math::

       \mathbf{p}=\frac{1}{V}\left(
       \sum_{i=1}^N m_i\mathbf{v}_i^\mathsf{T}\mathbf{v}_i
       +\mathbf{h}^\mathsf{T}\frac{dU}{d\mathbf{h}}\right)

    where :math:`V` is the volume, :math:`m_i` and :math:`\mathbf{v}_i`
    are the mass and velocity vector of particle :math:`i`,
    :math:`\mathbf{h}` is a :math:`3\times 3` matrix where the rows
    contain the box vectors, and :math:`U` is the total pairwise
    potential energy.

    To evaluate :math:`dU/d\mathbf{h}`, the box vectors are perturbed by
    `dh` in each of the six "directions" and the positions are updated
    accordingly. Then, OpenMM reevaluates the potential
    energy based on the new periodic box vectors and particle positions,
    and the difference in potential energy is used in the central finite
    difference formula to estimate the derivative.

    Parameters
    ----------
    context : `openmm.Context`
        OpenMM simulation context.

    dh : `float`, default: :code:`1e-5`
        Finite difference step size.

    diag : `bool`, keyword-only, default: :code:`False`
        Determines whether only the values in the main diagonal of the
        pressure tensor (i.e., the values that, when summed, gives the
        system pressure) are calculated.

    Returns
    -------
    pres : `numpy.ndarray`
        Estimated pressure tensor (or diagonal components).

        **Shape**: :math:`(3,)` or :math:`(3,\,3)`.

        **Reference unit**: :math:`\mathrm{atm}`.
    """

    try:
        state = context.getState(getPositions=True, getVelocities=True, getEnergy=True)
        box = state.getPeriodicBoxVectors(asNumpy=True)
        positions = state.getPositions(asNumpy=True)
        velocities = state.getVelocities(asNumpy=True)
        volume = box[0, 0] * box[1, 1] * box[2, 2]
    except openmm.OpenMMException:
        emsg = (
            "The simulation context must have information about "
            "the particle positions and velocities."
        )
        raise ValueError(emsg)

    system = context.getSystem()
    masses = (
        np.fromiter(
            (
                system.getParticleMass(i).value_in_unit(unit.dalton)
                for i in range(system.getNumParticles())
            ),
            dtype=float,
        )
        * unit.dalton
    )

    if diag:

        # Compute the ideal contribution
        p_kinetic = (masses * velocities**2).sum(axis=0)

        # Estimate the virial contribution with a central finite difference
        p_virial = np.zeros(3) * unit.kilojoule_per_mole
        for i in range(3):
            box_ = box.copy()
            box_[i, i] += dh
            context.setPeriodicBoxVectors(*box_)
            context.setPositions(
                np.dot(
                    positions,
                    np.divide(
                        box_,
                        box,
                        out=np.zeros_like(box),
                        where=box.value_in_unit(unit.nanometer) != 0,
                    ),
                )
            )
            U_plus = context.getState(getEnergy=True).getPotentialEnergy()
            box_ = box.copy()
            box_[i, i] -= dh
            context.setPeriodicBoxVectors(*box_)
            context.setPositions(
                np.dot(
                    positions,
                    np.divide(
                        box_,
                        box,
                        out=np.zeros_like(box),
                        where=box.value_in_unit(unit.nanometer) != 0,
                    ),
                )
            )
            U_minus = context.getState(getEnergy=True).getPotentialEnergy()
            p_virial[i] = U_plus - U_minus
        p_virial = (p_virial / (2 * dh)).in_units_of(p_kinetic.unit)

    else:

        # Compute the ideal contribution
        p_kinetic = (masses * velocities * velocities[:, :, None]).sum(axis=0)

        # Estimate the virial contribution with a central finite difference
        p_virial = np.zeros((3, 3)) * unit.kilojoule_per_mole
        for i in range(3):
            for j in range(i + 1):
                box_ = box.copy()
                box_[i, j] += dh
                context.setPeriodicBoxVectors(*box_)
                context.setPositions(
                    np.dot(
                        positions,
                        np.divide(
                            box_,
                            box,
                            out=np.zeros_like(box),
                            where=box.value_in_unit(unit.nanometer) != 0,
                        ),
                    )
                )
                U_plus = context.getState(getEnergy=True).getPotentialEnergy()
                box_ = box.copy()
                box_[i, j] -= dh
                context.setPeriodicBoxVectors(*box_)
                context.setPositions(
                    np.dot(
                        positions,
                        np.divide(
                            box_,
                            box,
                            out=np.zeros_like(box),
                            where=box.value_in_unit(unit.nanometer) != 0,
                        ),
                    )
                )
                U_minus = context.getState(getEnergy=True).getPotentialEnergy()
                p_virial[i, j] = U_plus - U_minus
        p_virial = (p_virial / (2 * dh)).in_units_of(p_kinetic.unit)
        p_virial = (
            p_virial._value + np.tril(p_virial).T - np.diag(np.diag(p_virial))
        ) * p_virial.unit

    return ((p_kinetic + p_virial) / (unit.AVOGADRO_CONSTANT_NA * volume)).in_units_of(
        unit.atmosphere
    )



"""
Topology and trajectory readers
===============================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains topology and trajectory readers for use with MDAnalysis.
"""

import concurrent.futures
import os
from pathlib import Path
from typing import TextIO, Union
import warnings

import psutil
from MDAnalysis.coordinates.base import ReaderBase
from MDAnalysis.lib.distances import transform_StoR
from MDAnalysis.lib.util import anyopen, store_init_arguments
import numpy as np

from ..algorithm.topology import convert_cell_representation


class LAMMPSDumpTrajectoryReader(ReaderBase):
    r"""
    LAMMPS dump trajectory reader.

    Parameters
    ----------
    filename : `str`
        Filename of the LAMMPS dump file.

    conventions : `str` or `tuple`, optional
        Coordinate conventions for the :math:`x`, :math:`y`, and
        :math:`z`-positions. If a `str` is provided, the same value is
        used for all axes. Determined automatically if not provided.

        .. container::

           **Valid values:**

           * :code:`None` to ignore an axis (when coordinate information
             is unavailable in that axis).
           * :code:`""` for unscaled and wrapped coordinates.
           * :code:`"u"` for unscaled and unwrapped coordinates.
           * :code:`"s"` for scaled and wrapped coordinates.
           * :code:`"su"` for scaled and unwrapped coordinates.

    unwrap : `bool`, default: :code:`False`
        Determines whether atom positions are unwrapped.

    extras : `list`, optional
        Extra per-atom information to be read from the dump file.
        Topology attributes, like type, mass, and charge, are not
        expected to change and are not supported.

        .. note::

           For named LAMMPS vector attributes, the stored array will
           always be arrays of shape :math:`(N,\,3)`, where :math:`N` is
           the number of atoms. If data for an axis is not found in the
           dump file, the corresponding column in the array will be
           filled with zeros.

           For custom LAMMPS vector attributes, the stored array will
           have as many columns as there are instances of the attribute
           in the dump file. For example, if :code:`"c_custom"` was
           provided to `extras` and the dump file contains the
           instances :code:`c_custom[2]`, :code:`c_custom[4]`, and
           :code:`c_custom[1]`, in that order, the array will have shape
           :math:`(N,\,3)`. The instances are always sorted, such that
           the columns in the example array correspond to
           :code:`c_custom[1]`, :code:`c_custom[2]`, and
           :code:`c_custom[4]`, respectively.

        **Valid values:**

        ==================================  ===================================================  ======================
        Keyword (`Reader.data` key)         LAMMPS dump attribute(s)                             Type
        ==================================  ===================================================  ======================
        :code:`"dipole_moments"`            (:code:`mux`, :code:`muy`, :code:`muz`)              `numpy.ndarray[float]`
        :code:`"dipole_moment_magnitudes"`  :code:`mu`                                           `float`
        :code:`"angular_velocities"`        (:code:`omegax`, :code:`omegay`, :code:`omegaz`)     `numpy.ndarray[float]`
        :code:`"angular_momentums"`         (:code:`angmomx`, :code:`angmomy`, :code:`angmomz`)  `numpy.ndarray[float]`
        :code:`"torques"`                   (:code:`tqx`, :code:`tqy`, :code:`tqz`)              `numpy.ndarray[float]`
        :code:`"c_{compute_id}"`            (:code:`c_{compute_id}[i]`, ...)                     `numpy.ndarray[float]`
        :code:`"d_{name}"`                  (:code:`d_{name}[i]`, ...)                           `numpy.ndarray[float]`
        :code:`"d2_{name}[i]"`              (:code:`d2_{name}[i][j]`, ...)                       `numpy.ndarray[float]`
        :code:`"f_{fix_id}"`                (:code:`f_{fix_id}[i]`, ...)                         `numpy.ndarray[float]`
        :code:`"i_{name}"`                  (:code:`i_{name}[i]`, ...)                           `numpy.ndarray[int]`
        :code:`"i2_{name}[i]"`              (:code:`i2_{name}[i][j]`, ...)                       `numpy.ndarray[int]`
        :code:`"v_{name}"`                  (:code:`v_{name}[i]`, ...)                           `numpy.ndarray[float]`
        ==================================  ===================================================  ======================

    Examples
    --------
    First, this trajectory reader must be registered to MDAnalysis by
    importing this submodule or this class:

    >>> from mdcraft.analysis import reader

    This will overwrite the built-in MDAnalysis LAMMPS dump reader.

    Then, to read a LAMMPS dump file :code:`simulation.lammpsdump` and
    extract both the topology and trajectory:

    >>> universe = mda.Universe("simulation.lammpsdump")

    If the dump file contains extra information, like the per-atom
    dipole moments, it can be specified in the :code:`extras` argument:

    >>> universe = mda.Universe("simulation.lammpsdump", extras=["dipole_moments"])

    The extra information will be stored in the :code:`data` attribute of
    the trajectory object:

    >>> dipole_moments = universe.trajectory.data["dipole_moments"]

    If the dump file does not have the :code:`.lammpsdump` extension,
    the format can be specified in the :code:`format` argument:

    >>> universe = mda.Universe("simulation.dump", format="LAMMPSDUMP")
    """

    _CONVENTIONS = ["u", "su", "", "s"]
    _CUSTOM_ATTRIBUTE_PREFIXES = ("c_", "d_", "d2_", "f_", "i_", "i2_", "v_")
    _EXTRA_ATTRIBUTES = {
        "dipole_moments": ("mux", "muy", "muz"),
        "dipole_moment_magnitudes": ("mu",),
        "angular_velocities": ("omegax", "omegay", "omegaz"),
        "angular_momentums": ("angmomx", "angmomy", "angmomz"),
        "torques": ("tqx", "tqy", "tqz"),
    }

    format = "LAMMPSDUMP"

    @staticmethod
    def _get_offsets(
        file: Union[str, Path, TextIO],
        start: int,
        end: int,
        grid: bool,
        parallel: bool,
    ) -> list[int]:
        """
        Finds the byte offsets for frames in a LAMMPS dump file.

        Parameters
        ----------
        file : `str`, `pathlib.Path`, or `io.TextIO`
            Filename, path, or handle to the LAMMPS dump file.

        start : `int`
            Byte offset to start reading from.

        end : `int`
            Byte offset to stop reading at.

        grid : `bool`
            Specifies whether the file uses the :code:`style grid`
            format.

        parallel : `bool`
            Determines whether the file is read in parallel.
        """

        manual = isinstance(file, (str, Path))
        if manual:
            file = open(file, "r")
        byte_counter = file.seek(start)

        offsets = []
        if parallel:
            while byte_counter < end:
                line = file.readline()
                byte_counter += len(line)
                if (line_index := line.find("ITEM: TIMESTEP")) != -1:
                    offsets.append(byte_counter - len(line) + line_index)
                    break
            while byte_counter < end:
                byte_counter += len(file.readline())  # <timestep>
                byte_counter += len(file.readline())  # ITEM: NUMBER OF ATOMS
                n_atoms = file.readline()
                byte_counter += len(n_atoms)
                for _ in range(5 + 4 * grid + int(n_atoms)):
                    byte_counter += len(file.readline())
                    if byte_counter >= end:
                        break
                else:
                    offsets.append(byte_counter)
                byte_counter += len(file.readline())  # ITEM: TIMESTEP
        else:
            line_counter = 0
            lines_per_frame = 9 + 4 * grid
            line = True
            while line:
                if (relative_line := line_counter % lines_per_frame) == 0 and (
                    offset := file.tell()
                ) < end:
                    offsets.append(offset)
                elif relative_line == 4:
                    lines_per_frame = 9 + 4 * grid + int(line)
                line = file.readline()
                line_counter += 1

        if manual:
            file.close()

        return offsets

    @store_init_arguments
    def __init__(
        self,
        filename: str,
        conventions: Union[str, tuple[str]] = None,
        unwrap: bool = False,
        *,
        extras: list[str] = None,
        parallel: bool = False,
        n_threads: int = None,
        **kwargs,
    ) -> None:

        super().__init__(filename, **kwargs)

        if conventions is not None:
            if isinstance(conventions, str):
                if conventions not in self._CONVENTIONS:
                    emsg = (
                        f"Invalid convention '{conventions}'. Valid "
                        "values: '" + "', '".join(self._CONVENTIONS) + "'."
                    )
                    raise ValueError(emsg)
                self._conventions = 3 * [conventions]
            elif len(conventions) != 3:
                emsg = "'conventions' must have length 3 when it is an array."
                raise ValueError(emsg)
            else:
                for c in conventions:
                    if c not in self._CONVENTIONS:
                        emsg = (
                            f"Invalid convention '{c}'. Valid values: "
                            "'" + "', '".join(self._CONVENTIONS) + "'."
                        )
                        raise ValueError(emsg)
                self._conventions = conventions
        else:
            self._conventions = []

        with anyopen(self.filename) as f:
            for _ in range(3):
                f.readline()
            self.n_atoms = int(f.readline())

            for _ in range(4):
                f.readline()
            self.is_style_grid = f.readline().rstrip() == "ITEM: DIMENSION"

        file_size = os.path.getsize(filename)
        if parallel:
            n_threads = n_threads or psutil.cpu_count()
            chunk_size = np.ceil(file_size / n_threads).astype(int)
            self._offsets = []
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=n_threads
            ) as executor:
                for future in concurrent.futures.as_completed(
                    executor.submit(
                        self._get_offsets,
                        filename,
                        i * chunk_size,
                        min((i + 1) * chunk_size, file_size),
                        self.is_style_grid,
                        True,
                    )
                    for i in range(n_threads)
                ):
                    self._offsets.extend(future.result())
            self._offsets.sort()
        else:
            self._file = open(filename, "r")
            self._offsets = self._get_offsets(
                self._file, 0, file_size, self.is_style_grid, False
            )

        if isinstance(extras, str):
            if extras not in self._EXTRA_ATTRIBUTES and not extras.startswith(
                self._CUSTOM_ATTRIBUTE_PREFIXES
            ):
                raise ValueError(f"Invalid attribute '{extras}' in 'extras'.")
            extras = [extras]
        elif extras is not None:
            for attr in extras:
                if attr not in self._EXTRA_ATTRIBUTES and not attr.startswith(
                    self._CUSTOM_ATTRIBUTE_PREFIXES
                ):
                    raise ValueError(f"Invalid attribute '{attr}' in 'extras'.")
        self._extras = extras

        self._unwrap = unwrap

        self._reopen()
        self._read_next_timestep()

    def _reopen(self) -> None:
        """
        Reopens the LAMMPS dump file.
        """

        self.close()
        self._file = anyopen(self.filename)
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        self.ts.frame = -1

    def _read_frame(self, frame: int) -> ReaderBase._Timestep:
        """
        Reads a specific frame from the LAMMPS dump file.

        Parameters
        ----------
        frame : `int`, optional
            Frame to read.

        Returns
        -------
        timestep : `ReaderBase._Timestep`
            Timestep object with all information from a frame in the
            trajectory.
        """

        self._file.seek(self._offsets[frame])
        self.ts.frame = frame - 1
        return self._read_next_timestep()

    def _read_next_timestep(self, ts: int = None) -> ReaderBase._Timestep:
        """
        Reads the next timestep from the LAMMPS dump file.

        Returns
        -------
        timestep : `ReaderBase._Timestep`
            Timestep object with all information from a frame in the
            trajectory.
        """

        # Set up timestep
        ts = self.ts
        ts.frame += 1
        if ts.frame >= self.n_frames:
            emsg = (
                f"'{self.filename}' only has {self.n_frames} frames "
                f"and does not contain a frame {ts.frame}."
            )
            raise EOFError(emsg)

        # Get file and read timestep information
        f = self._file
        f.readline()  # ITEM: TIMESTEP
        ts.data["step"] = int(f.readline())
        ts.data["time"] = ts.data["step"] * ts.dt

        # Read number of atoms
        f.readline()  # ITEM: NUMBER OF ATOMS
        if (n_atoms := int(f.readline())) != self.n_atoms:
            emsg = (
                f"Timestep {ts.data['step']} has {n_atoms} atoms, "
                f"but the topology has {self.n_atoms} atoms."
            )
            raise RuntimeError(emsg)

        # Read box information
        if "xy xz yz" in (_ := f.readline()):  # ITEM: BOX BOUNDS
            xlo, xhi, xy = (float(v) for v in f.readline().split())
            ylo, yhi, xz = (float(v) for v in f.readline().split())
            zlo, zhi, yz = (float(v) for v in f.readline().split())

            # https://docs.lammps.org/Howto_triclinic.html\
            # #output-of-restricted-and-general-triclinic-boxes-in-a-dump-file
            xlo -= min(0, xy, xz, xy + xz)
            xhi -= max(0, xy, xz, xy + xz)
            ylo -= min(0, yz)
            yhi -= max(0, yz)

            box_vectors = np.zeros((3, 3))
            box_vectors[0, 0] = xhi - xlo
            box_vectors[1, :2] = xy, yhi - ylo
            box_vectors[2] = xz, yz, zhi - zlo
            ts.dimensions = convert_cell_representation(vectors=box_vectors)
        elif "abc origin" in _:
            box_vectors = np.empty((3, 3))
            for i in range(3):
                box_vectors[i] = (float(v) for v in f.readline().split()[:3])
            ts.dimensions = convert_cell_representation(vectors=box_vectors)
        else:
            xlo, xhi = (float(v) for v in f.readline().split())
            ylo, yhi = (float(v) for v in f.readline().split())
            zlo, zhi = (float(v) for v in f.readline().split())
            ts.dimensions = np.array((xhi - xlo, yhi - ylo, zhi - zlo, 90, 90, 90))

        if self.is_style_grid:
            for _ in range(4):
                f.readline()

        # Get all attributes in dump file
        attributes = {a: i for i, a in enumerate(f.readline().rstrip().split()[2:])}

        # Get coordinate axis indices and verify conventions
        self._axis_indices = np.empty(3, dtype=int)
        if self._conventions:
            for i, (ax, c) in enumerate(zip("xyz", self._conventions)):
                if c is None:
                    self._axis_indices[i] = -1
                else:
                    if (col := attributes.get(f"{ax}{c}")) is None:
                        emsg = (
                            f"No {ax}-coordinate information with the "
                            f"specified convention '{c}' found in "
                            f"'{self.filename}'."
                        )
                        raise RuntimeError(emsg)
                    self._axis_indices[i] = col
        else:
            for i, ax in enumerate("xyz"):
                for c in self._CONVENTIONS:
                    if (n := f"{ax}{c}") in attributes:
                        self._axis_indices[i] = attributes[n]
                        self._conventions.append(c)
                        break
                else:
                    self._axis_indices[i] = -1
                    self._conventions.append(None)
        self._scaled_where = np.asarray(
            [i for i, c in enumerate(self._conventions) if "s" in c], dtype=int
        )
        self._has_scaled = np.any(self._scaled_where)

        # Get image flag indices if unwrapping is desired
        if self._unwrap:
            self._image_indices = np.fromiter(
                (
                    -1 if ai == -1 or "u" in c else (attributes.get(f"i{ax}") or -1)
                    for ax, ai, c in zip("xyz", self._axis_indices, self._conventions)
                ),
                dtype=int,
                count=3,
            )
            if np.any(self._image_indices >= 0):
                self._image_where = np.where(self._image_indices >= 0)[0]
            else:
                wmsg = "No image flags found in dump file. " "Setting 'unwrap=False'."
                warnings.warn(wmsg)
                self._unwrap = False

        # Check if velocity and force information is available
        self._velocity_indices = np.fromiter(
            (attributes.get(f"v{ax}") or -1 for ax in "xyz"), dtype=int, count=3
        )
        ts.has_velocities = np.any(self._velocity_indices)
        self._force_indices = np.fromiter(
            (attributes.get(f"f{ax}") or -1 for ax in "xyz"), dtype=int, count=3
        )
        ts.has_forces = np.any(self._force_indices)

        # Preallocate arrays for extra information, if requested
        if self._extras:
            self._extras_indices = {}
            for attr in self._extras:
                if attr.startswith(self._CUSTOM_ATTRIBUTE_PREFIXES):
                    la = len(attr)
                    self._extras_indices[attr] = [
                        j
                        for _, j in sorted(
                            (int(a[a.find("[", la) + 1 : a.find("]", la)]), i)
                            for a, i in attributes.items()
                            if a.startswith(attr)
                        )
                    ]
                    ts.data[attr] = np.empty(
                        (n_atoms, len(self._extras_indices[attr])),
                        dtype=np.int64 if attr.startswith("i_") else np.float32,
                    )
                else:
                    self._extras_indices[attr] = [
                        attributes.get(a) or -1 for a in self._EXTRA_ATTRIBUTES[attr]
                    ]
                    ts.data[attr] = np.empty(
                        (n_atoms, len(self._extras_indices[attr])), dtype=np.float32
                    )

        # Iterate over atoms and read their information
        if has_id := "id" in attributes:
            atom_indices = np.empty(self.n_atoms, dtype=int)
        for i in range(self.n_atoms):
            values = f.readline().split()

            # Store atom ID
            if has_id:
                atom_indices[i] = values[attributes["id"]]

            # Store atom positions (missing dimensions are left as 0)
            ts.positions[i] = tuple(
                0 if ai == -1 else values[ai] for ai in self._axis_indices
            )

            # Unscale coordinates before unwrapping, if necessary
            if self._has_scaled:
                ts.positions[i, self._scaled_where] = transform_StoR(
                    ts.positions[i], ts.dimensions
                )[self._scaled_where]

            # Unwrap coordinates, if necessary
            if self._unwrap:
                ts.positions[i, self._image_where] += ts.dimensions[
                    self._image_where
                ] * tuple(
                    int(values[ii]) for ii in self._image_indices[self._image_where]
                )

            # Shift unwrapped and/or unscaled coordinates by the origin
            if self._has_scaled or self._unwrap:
                ts.positions += (xlo, ylo, zlo)

            # Store atom velocities and forces, if available
            if ts.has_velocities:
                ts.velocities[i] = tuple(
                    0 if vi == -1 else values[vi] for vi in self._velocity_indices
                )
            if ts.has_forces:
                ts.forces[i] = tuple(
                    0 if fi == -1 else values[fi] for fi in self._force_indices
                )

            # Store extra information, if requested
            if self._extras:
                for attr, attribute_indices in self._extras_indices.items():
                    if attr.startswith(self._CUSTOM_ATTRIBUTE_PREFIXES):
                        ts.data[attr][i] = np.fromiter(
                            (values[ai] for ai in attribute_indices),
                            dtype=np.int64 if attr.startswith("i_") else np.float32,
                            count=len(attribute_indices),
                        )
                    else:
                        ts.data[attr][i] = tuple(
                            0 if ai == -1 else float(values[ai])
                            for ai in attribute_indices
                        )

        # Reorder atoms by ID, if available
        if has_id:
            order = np.argsort(atom_indices)
            ts.positions = ts.positions[order]
            if ts.has_velocities:
                ts.velocities = ts.velocities[order]
            if ts.has_forces:
                ts.forces = ts.forces[order]
            if self._extras:
                for attr in self._extras:
                    ts.data[attr] = ts.data[attr][order]

        return ts

    @property
    def n_frames(self) -> int:
        return len(self._offsets)

    def close(self) -> None:
        """
        Closes the LAMMPS dump file.
        """

        if hasattr(self, "_file"):
            self._file.close()



"""
Custom OpenMM topology and trajectory file readers and writers
==============================================================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module provides custom topology and trajectory file readers and
writers for OpenMM.
"""

import platform
from typing import Any, Union
import warnings

import netCDF4 as nc
import numpy as np
import openmm
from openmm import app, unit

from .. import VERSION


class NetCDFFile:
    """
    Interface for reading and writing AMBER NetCDF trajectory and
    restart files.

    Parameters
    ----------
    file : `str` or `netcdf4.Dataset`
        NetCDF file. If `file` is a filename and does not have the
        :code:`.nc` or :code:`.ncdf` extension, :code:`.nc` will
        automatically be appended.

    mode : `str`
        NetCDF file access mode.

    restart : `bool`, default: :code:`False`
        Specifies whether the NetCDF file is a trajectory or restart file.

    **kwargs
        Keyword arguments to be passed to :code:`netCDF4.Dataset`.
    """

    def __init__(
        self, file: Union[str, nc.Dataset], mode: str, restart: bool = False, **kwargs
    ):

        if isinstance(file, str):
            if not file.endswith((".nc", ".ncdf")):
                file += ".nc"
            self._nc = nc.Dataset(
                file, mode=mode, format="NETCDF3_64BIT_OFFSET", **kwargs
            )
        else:
            self._nc = file
        self._nc.set_always_mask(False)

        if mode.startswith(("a", "r+", "w", "x")):
            self._restart = restart
        else:
            self._restart = self._nc.Conventions == "AMBERRESTART"
        self._frame = self.get_num_frames() if hasattr(self._nc, "Conventions") else 0

    def get_dimensions(
        self, frames: Union[int, list[int], slice] = None, units: bool = True
    ) -> Union[
        tuple[np.ndarray[float], np.ndarray[float]], tuple[unit.Quantity, unit.Quantity]
    ]:
        """
        Get the simulation box dimensions.

        Parameters
        ----------
        frames : `int`, `list`, or `slice`, optional
            Frame indices. If :code:`None`, the dimensions across all
            frames are returned.

        units : `bool`, default: :code:`True`
            Determines whether the dimensions are returned with units.

        Returns
        -------
        cell_lengths : `numpy.ndarray`, optional
            Simulation box dimensions.

            **Reference unit**: :math:`\\mathrm{Å}`.

        cell_angles : `numpy.ndarray`, optional
            Angles that define the shape of the simulation box.

            **Reference unit**: :math:`^\\circ`.
        """

        cell_lengths = (
            self._nc.variables["cell_lengths"][:]
            if frames is None
            else self._nc.variables["cell_lengths"][frames]
        )
        cell_angles = (
            self._nc.variables["cell_angles"][:]
            if frames is None
            else self._nc.variables["cell_angles"][frames]
        )
        if units:
            cell_lengths *= unit.angstrom
            cell_angles *= unit.degree
        return cell_lengths, cell_angles

    def get_num_frames(self) -> int:
        """
        Get the number of frames.

        Returns
        -------
        num_frames : `int`
            Number of frames.
        """

        if not hasattr(self._nc, "Conventions"):
            raise RuntimeError(
                "The NetCDF file is not a valid AMBER NetCDF "
                "trajectory or does not contain any data."
            )
        return self._nc.dimensions["frame"].size

    def get_num_atoms(self) -> int:
        """
        Get the number of atoms.

        Returns
        -------
        num_atoms : `int`
            Number of atoms.
        """

        if not hasattr(self._nc, "Conventions"):
            raise RuntimeError(
                "The NetCDF file is not a valid AMBER NetCDF "
                "trajectory or does not contain any data."
            )
        return self._nc.dimensions["atom"].size if hasattr(self._nc, "Conventions") else 0

    def get_times(
        self, frames: Union[int, list[int], slice] = None, units: bool = True
    ) -> Union[np.ndarray[float], unit.Quantity]:
        """
        Get simulation times.

        Parameters
        ----------
        frames : `int`, `list`, or `slice`, optional
            Frame indices. If :code:`None`, the times across all
            frames are returned.

        units : `bool`, default: :code:`True`
            Determines whether the times are returned with units.

        Returns
        -------
        times : `numpy.ndarray` or `openmm.unit.Quantity`
            Simulation times.

            **Reference unit**: :math:`\\mathrm{ps}`.
        """

        if not hasattr(self._nc, "Conventions"):
            raise RuntimeError(
                "The NetCDF file is not a valid AMBER NetCDF "
                "trajectory or does not contain any data."
            )
        times = (
            self._nc.variables["time"][:]
            if frames is None
            else self._nc.variables["time"][frames]
        )
        if units:
            times *= unit.picosecond
        return times

    def get_positions(
        self, frames: Union[int, list[int], slice] = None, units: bool = True
    ) -> Union[np.ndarray[float], unit.Quantity]:
        """
        Get the atom positions.

        Parameters
        ----------
        frames : `int`, `list`, or `slice`, optional
            Frame indices. If :code:`None`, the positions across all
            frames are returned.

        units : `bool`, default: :code:`True`
            Determines whether the positions are returned with units.

        Returns
        -------
        positions : `numpy.ndarray` or `openmm.unit.Quantity`
            Atom positions.

            **Reference unit**: :math:`\\mathrm{Å}`.
        """

        if not hasattr(self._nc, "Conventions"):
            raise RuntimeError(
                "The NetCDF file is not a valid AMBER NetCDF "
                "trajectory or does not contain any data."
            )
        positions = (
            self._nc.variables["coordinates"][:]
            if frames is None
            else self._nc.variables["coordinates"][frames]
        )
        if units:
            positions *= unit.angstrom
        return positions

    def get_velocities(
        self, frames: Union[int, list[int], slice] = None, units: bool = True
    ) -> Union[np.ndarray[float], unit.Quantity]:
        """
        Get atom velocities.

        Parameters
        ----------
        frames : `int`, `list`, or `slice`, optional
            Frame indices. If :code:`None`, the velocities across all
            frames are returned.

        units : `bool`, default: :code:`True`
            Determines whether the velocities are returned with units.

        Returns
        -------
        velocities : `numpy.ndarray` or `openmm.unit.Quantity`
            Atom velocities. If the NetCDF file does not contain
            this information, :code:`None` is returned.

            **Reference unit**: :math:`\\mathrm{Å/ps}`.
        """

        if not hasattr(self._nc, "Conventions"):
            raise RuntimeError(
                "The NetCDF file is not a valid AMBER NetCDF "
                "trajectory or does not contain any data."
            )
        if "velocities" not in self._nc.variables:
            wmsg = (
                "The NetCDF file does not contain information about "
                "the atom velocities."
            )
            warnings.warn(wmsg)
            return None

        velocities = (
            self._nc.variables["velocities"][:]
            if frames is None
            else self._nc.variables["velocities"][frames]
        )
        if units:
            velocities *= unit.angstrom / unit.picosecond
        return velocities

    def get_forces(
        self, frames: Union[int, list[int], slice] = None, units: bool = True
    ) -> Union[np.ndarray[float], unit.Quantity]:
        """
        Get the forces acting on the atoms.

        Parameters
        ----------
        frames : `int`, `list`, or `slice`, optional
            Frame indices. If :code:`None`, the forces across all frames
            are returned.

        units : `bool`, default: :code:`True`
            Determines whether the forces are returned with units.

        Returns
        -------
        forces : `numpy.ndarray` or `openmm.unit.Quantity`
            Forces acting on the atoms. If the NetCDF file does not
            contain this information, :code:`None` is returned.

            **Reference unit**: :math:`\\mathrm{Å/ps}`.
        """

        if not hasattr(self._nc, "Conventions"):
            raise RuntimeError(
                "The NetCDF file is not a valid AMBER NetCDF "
                "trajectory or does not contain any data."
            )
        if "forces" not in self._nc.variables:
            wmsg = (
                "The NetCDF file does not contain information about "
                "the forces acting on the atoms."
            )
            warnings.warn(wmsg)
            return None

        forces = (
            self._nc.variables["forces"][:]
            if frames is None
            else self._nc.variables["forces"][frames]
        )
        if units:
            forces *= unit.kilocalorie_per_mole / unit.angstrom
        return forces

    def write_header(
        self: Any,
        N: int,
        cell: bool,
        velocities: bool,
        forces: bool,
        restart: bool = False,
        *,
        remd: str = None,
        temp0: float = None,
        remd_dimtype: np.ndarray[int] = None,
        remd_indices: np.ndarray[int] = None,
        remd_repidx: int = -1,
        remd_crdidx: int = -1,
        remd_values: np.ndarray[float] = None,
    ) -> "NetCDFFile":
        """
        Initialize a NetCDF file according to `AMBER NetCDF
        Trajectory/Restart Convention Version 1.0, Revision C
        <https://ambermd.org/netcdf/nctraj.xhtml>`_.

        .. note::

           This function can be used as either a static or instance
           method.

        Parameters
        ----------
        self : `str`, `netcdf4.Dataset`, or `mdcraft.openmm.file.NetCDFFile`
            If :meth:`write_header` is called as a static method, you
            must provide a filename or a NetCDF file object. Otherwise,
            the NetCDF file embedded in the current instance is used.

        N : `int`
            Number of atoms.

        cell : `bool`
            Specifies whether simulation box length and angle
            information is available.

        velocities : `bool`
            Specifies whether atom velocities should be written.

        forces : `bool`
            Specifies whether forces exerted on atoms should be
            written.

        restart : `bool`, default: :code:`False`
            Specifies whether the NetCDF file is a trajectory or restart
            file.

        remd : `str`, keyword-only, optional
            Specifies whether information about a replica exchange
            molecular dynamics (REMD) simulation is written.

            .. container::

               **Valid values**:

               * :code:`"temp"` for regular REMD.
               * :code:`"multi"` for multi-dimensional REMD.

        temp0 : `float`, keyword-only, optional
            Temperature that the thermostat is set to maintain for a
            REMD restart file only.

            **Reference unit**: :math:`\\mathrm{K}`.

        remd_dimtype : array-like, keyword-only, optional
            Array specifying the exchange type(s) for the REMD
            dimension(s). Required for a multi-dimensional REMD restart
            file.

        remd_indices : array-like, keyword-only, optional
            Array specifying the position in all dimensions that each
            frame is in. Required for a multi-dimensional REMD restart
            file.

        remd_repidx : `int`, keyword-only, optional
            Overall index of the frame in replica space.

        remd_crdidx : `int`, keyword-only, optional
            Overall index of the frame in coordinate space.

        remd_values : array-like, keyword-only, optional
            Replica value the specified replica dimension has for that
            given frame. Required for a multi-dimensional REMD restart
            file.

        Returns
        -------
        netcdf_file : `mdcraft.openmm.file.NetCDFFile`
            NetCDF file object. Only returned when this function is used
            as a static method.
        """

        # Create NetCDF object if it doesn't already exist
        if not isinstance(self, NetCDFFile):
            self = NetCDFFile(self, "w", restart=restart)

        self._nc.Conventions = "AMBER"
        if self._restart:
            self._nc.Conventions += "RESTART"
        self._nc.ConventionVersion = "1.0"
        self._nc.program = "MDCraft"
        self._nc.programVersion = VERSION
        self._nc.title = (
            f"OpenMM {openmm.Platform.getOpenMMVersion()} / " f"{platform.node()}"
        )

        if self._restart:
            self._nc.createDimension("frame", 1)
        else:
            self._nc.createDimension("frame", None)

        if remd == "multi":  # pragma: no cover
            self._nc.createDimension("remd_dimension", len(remd_dimtype))
        self._nc.createDimension("spatial", 3)
        self._nc.createDimension("atom", N)

        if self._restart:
            self._nc.createVariable("coordinates", "d", ("atom", "spatial"))
        else:
            self._nc.createVariable("coordinates", "f", ("frame", "atom", "spatial"))
        self._nc.variables["coordinates"].units = "angstrom"

        self._nc.createVariable("time", "d", ("frame",))
        self._nc.variables["time"].units = "picosecond"

        if cell:
            self._nc.createDimension("cell_spatial", 3)
            self._nc.createDimension("cell_angular", 3)
            self._nc.createDimension("label", 5)
            self._nc.createVariable("spatial", "c", ("spatial",))
            self._nc.variables["spatial"][:] = list("xyz")
            self._nc.createVariable("cell_spatial", "c", ("cell_spatial",))
            self._nc.variables["cell_spatial"][:] = list("abc")
            self._nc.createVariable("cell_angular", "c", ("cell_angular", "label"))
            self._nc.variables["cell_angular"][:] = [
                list("alpha"),
                list("beta "),
                list("gamma"),
            ]

            if self._restart:
                self._nc.createVariable("cell_lengths", "d", ("cell_spatial",))
                self._nc.createVariable("cell_angles", "d", ("cell_angular",))
            else:
                self._nc.createVariable("cell_lengths", "f", ("frame", "cell_spatial"))
                self._nc.createVariable("cell_angles", "f", ("frame", "cell_angular"))
            self._nc.variables["cell_lengths"].units = "angstrom"
            self._nc.variables["cell_angles"].units = "degree"

        if velocities:
            if self._restart:
                self._nc.createVariable("velocities", "d", ("atom", "spatial"))
            else:
                self._nc.createVariable("velocities", "f", ("frame", "atom", "spatial"))
            self._nc.variables["velocities"].units = "angstrom/picosecond"
            self._nc.variables["velocities"].scale_factor = 20.455

        if forces:
            if self._restart:
                self._nc.createVariable("forces", "d", ("atom", "spatial"))
            else:
                self._nc.createVariable("forces", "f", ("frame", "atom", "spatial"))
            self._nc.variables["forces"].units = "kilocalorie/mole/angstrom"

        if remd is not None:  # pragma: no cover
            if remd == "temp":
                self._nc.createVariable("temp0", "d", ("frame",))
                if self._restart:
                    if temp0 is None:
                        emsg = (
                            "Temperature must be provided for a REMD " "restart file."
                        )
                        raise ValueError(emsg)
                    self._nc.variables["temp0"][0] = temp0
                self._nc.variables["temp0"].units = "kelvin"

            elif remd == "multi":
                self._nc.createVariable("remd_dimtype", "i", ("remd_dimension",))
                self._nc.createVariable("remd_repidx", "i", ("frame",))
                self._nc.createVariable("remd_crdidx", "i", ("frame",))
                if self._restart:
                    if remd_dimtype is None:
                        emsg = (
                            "Dimension types must be provided for a "
                            "multi-dimensional REMD restart file."
                        )
                        raise ValueError(emsg)
                    self._nc.variables["remd_dimtype"] = remd_dimtype

                    self._nc.createVariable("remd_indices", "i", ("remd_dimension",))
                    if remd_indices is None:
                        emsg = (
                            "Dimension indices must be provided for a "
                            "multi-dimensional REMD restart file."
                        )
                        raise ValueError(emsg)
                    self._nc.variables["remd_indices"] = remd_indices

                    self._nc.variables["remd_repidx"][0] = remd_repidx
                    self._nc.variables["remd_crdidx"][0] = remd_crdidx

                    self._nc.createVariable("remd_values", "d", ("remd_dimension",))
                    if remd_values is None:
                        emsg = (
                            "Replica values must be provided for a "
                            "multi-dimensional REMD restart file."
                        )
                        raise ValueError(emsg)
                    self._nc.variables["remd_values"][:] = remd_values

                else:
                    self._nc.createVariable(
                        "remd_indices", "i", ("frame", "remd_dimension")
                    )
                    self._nc.createVariable(
                        "remd_values", "d", ("frame", "remd_dimension")
                    )

        return self

    def write_file(self: Any, state: openmm.State) -> "NetCDFFile":
        """
        Write a single simulation state to a restart NetCDF file.

        .. note::

           This function can be used as either a static or instance
           method.

        Parameters
        ----------
        self : `str`, `netcdf4.Dataset`, or `mdcraft.openmm.file.NetCDFFile`
            If :meth:`write_file` is called as a static method, you must
            provide a filename or a NetCDF file object. Otherwise, the
            NetCDF file embedded in the current instance is used.

        state : `openmm.State`
            OpenMM simulation state from which to retrieve cell
            dimensions and atom positions, velocities, and forces.

        Returns
        -------
        netcdf_file : `mdcraft.openmm.file.NetCDFFile`
            NetCDF file object. Only returned when this function is used
            as a static method.
        """

        # Collect all available data in the state
        data = {}
        pbv = state.getPeriodicBoxVectors()
        if pbv is not None:
            (a, b, c, alpha, beta, gamma) = (
                app.internal.unitcell.computeLengthsAndAngles(pbv)
            )
            data["cell_lengths"] = 10 * np.array((a, b, c))
            data["cell_angles"] = 180 * np.array((alpha, beta, gamma)) / np.pi
        data["coordinates"] = state.getPositions(asNumpy=True).value_in_unit(
            unit.angstrom
        )
        try:
            data["velocities"] = state.getVelocities(asNumpy=True).value_in_unit(
                unit.angstrom / unit.picosecond
            )
        except openmm.OpenMMException:  # pragma: no cover
            pass
        try:
            data["forces"] = state.getForces(asNumpy=True).value_in_unit(
                unit.kilocalorie_per_mole / unit.angstrom
            )
        except openmm.OpenMMException:  # pragma: no cover
            pass

        # Create NetCDF file or object if it doesn't already exist
        if not isinstance(self, NetCDFFile):
            self = NetCDFFile(self, "w", restart=True)
        if not hasattr(self._nc, "Conventions"):
            self.write_header(
                data["coordinates"].shape[0],
                "cell_lengths" in data or "cell_angles" in data,
                "velocities" in data,
                "forces" in data,
            )
            self._nc.set_always_mask(False)
        elif self._nc.Conventions != "AMBERRESTART":
            raise ValueError("The NetCDF file must be a restart file.")

        # Write data to NetCDF file
        for k, v in data.items():
            self._nc.variables[k][:] = v
        self._nc.sync()

        return self

    def write_model(
        self: Any,
        time: Union[float, np.ndarray[float]],
        coordinates: np.ndarray[float],
        velocities: np.ndarray[float] = None,
        forces: np.ndarray[float] = None,
        cell_lengths: np.ndarray[float] = None,
        cell_angles: np.ndarray[float] = None,
        *,
        restart: bool = False,
    ) -> "NetCDFFile":
        """
        Write simulation state(s) to a NetCDF file.

        .. note::

           This function can be used as either a static or instance
           method.

        Parameters
        ----------
        self : `str`, `netcdf4.Dataset`, or `mdcraft.openmm.file.NetCDFFile`
            If :meth:`write_model` is called as a static method, you must
            provide a filename or a NetCDF file object. Otherwise, the
            NetCDF file embedded in the current instance is used.

        time : `float` or `numpy.ndarray`
            Time(s). The dimensionality determines whether a single or
            multiple frames are written.

            **Reference unit**: :math:`\\mathrm{ps}`.

        coordinates : `numpy.ndarray`
            Coordinates of :math:`N` atoms over :math:`N_t` frames. The
            dimensionality depends on whether a single or multiple
            frames are to be written and must be compatible with that
            for `time`.

            **Shape**: :math:`(N,\\,3)` or :math:`(N_t,\\,N,\\,3)`.

            **Reference unit**: :math:`\\mathrm{Å}`.

        velocities : `numpy.ndarray`, optional
            Velocities of :math:`N` atoms over :math:`N_t` frames. The
            dimensionality depends on whether a single or multiple
            frames are to be written and must be compatible with that
            for `time`.

            **Shape**: :math:`(N,\\,3)` or :math:`(N_t,\\,N,\\,3)`.

            **Reference unit**: :math:`\\mathrm{Å/ps}`.

        forces : `numpy.ndarray`, optional
            Forces exerted on :math:`N` atoms over :math:`N_t` frames.
            The dimensionality depends on whether a single or multiple
            frames are to be written and must be compatible with that
            for `time`.

            **Shape**: :math:`(N,\\,3)` or :math:`(N_t,\\,N,\\,3)`.

            **Reference unit**: :math:`\\mathrm{Å/ps}`.

        cell_lengths : `numpy.ndarray`, optional
            Simulation box dimensions.

            **Shape**: :math:`(3,)`.

            **Reference unit**: :math:`\\mathrm{Å}`.

        cell_angles : `numpy.ndarray`, optional
            Angles that define the shape of the simulation box.

            **Shape**: :math:`(3,)`.

            **Reference unit**: :math:`^\\circ`.

        restart : `bool`, keyword-only, default: :code:`False`
            Prevents the frame index from being incremented if writing a
            NetCDF restart file.

        Returns
        -------
        netcdf_file : `mdcraft.openmm.file.NetCDFFile`
            NetCDF file object. Only returned when this function is used
            as a static method.
        """

        # Create NetCDF file or object if it doesn't already exist
        if not isinstance(self, NetCDFFile):
            self = NetCDFFile(self, "w", restart=restart)
        if not hasattr(self._nc, "Conventions"):
            self.write_header(
                coordinates.shape[0],
                cell_lengths is not None or cell_angles is not None,
                velocities is not None,
                forces is not None,
            )
            self._nc.set_always_mask(False)

        # Write model to NetCDF file
        n_frames = len(time) if isinstance(time, (tuple, list, np.ndarray)) else 1
        frames = slice(self._frame, self._frame + n_frames)
        self._nc.variables["time"][frames] = time
        self._nc.variables["coordinates"][frames] = coordinates
        if velocities is not None:
            self._nc.variables["velocities"][frames] = velocities
        if forces is not None:
            self._nc.variables["forces"][frames] = forces
        if cell_lengths is not None:
            self._nc.variables["cell_lengths"][frames] = cell_lengths
        if cell_angles is not None:
            self._nc.variables["cell_angles"][frames] = cell_angles
        self._nc.sync()
        if not restart:
            self._frame += n_frames

        return self



import pathlib
import sys

import numpy as np
from openmm import unit
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft import ureg
from mdcraft.algorithm.unit import get_lj_scale_factors, strip_unit # noqa: E402

def test_func_get_lj_scaling_factors():

    # TEST CASE 1: Lennard-Jones scaling factors
    pint_factors = get_lj_scale_factors({
        "mass": 39.948 * ureg.gram / ureg.mole,
        "energy": 3.9520829798737548e-25 * ureg.kilocalorie,
        "length": 3.4 * ureg.angstrom
    })
    openmm_factors = get_lj_scale_factors({
        "mass": 39.948 * unit.gram / unit.mole,
        "energy": 0.238 * unit.kilocalorie_per_mole / unit.AVOGADRO_CONSTANT_NA,
        "length": 3.4 * unit.angstrom
    })
    for key in openmm_factors.keys():
        value, unit_ = strip_unit(openmm_factors[key])
        assert np.isclose(strip_unit(pint_factors[key], unit_)[0],
                          value)

def test_func_strip_unit():

    # TEST CASE 1: Strip unit from non-Quantity
    assert strip_unit(90.0, "deg") == (90.0, "deg")
    assert strip_unit(90.0, ureg.degree) == (90.0, ureg.degree)
    assert strip_unit(90.0, unit.degree) == (90.0, unit.degree)

    # TEST CASE 2: Strip unit from Quantity
    k_ = 1.380649e-23
    assert strip_unit(k_) == (k_, None)
    assert strip_unit(k_ * ureg.joule * ureg.kelvin ** -1) \
           == (k_, ureg.joule * ureg.kelvin ** -1)
    assert strip_unit(k_ * unit.joule * unit.kelvin ** -1) \
           == (k_, unit.joule * unit.kelvin ** -1)

    # TEST CASE 3: Strip unit from Quantity with compatible unit specified
    g_ = 32.17404855643044
    g = 9.80665 * ureg.meter / ureg.second ** 2
    assert strip_unit(g_, "foot/second**2") \
           == (g_, ureg.foot / ureg.second ** 2)
    assert strip_unit(g, ureg.foot / ureg.second ** 2) \
           == (g_, ureg.foot / ureg.second ** 2)
    g = 9.80665 * unit.meter / unit.second ** 2
    assert strip_unit(g, "foot/second**2") \
           == (g_, unit.foot / unit.second ** 2)
    assert strip_unit(g, unit.foot / unit.second ** 2) \
           == (g_, unit.foot / unit.second ** 2)

    # TEST CASE 4: Strip unit from Quantity with incompatible unit specified
    R_ = 8.31446261815324
    R__ = 8.205736608095969e-05
    assert strip_unit(
        R__ * ureg.meter ** 3 * ureg.atmosphere / (ureg.kelvin * ureg.mole),
        unit.joule / (unit.kelvin * unit.mole)
    ) == (R_, unit.joule / (unit.kelvin * unit.mole))
    assert strip_unit(
        R__ * unit.meter ** 3 * unit.atmosphere / (unit.kelvin * unit.mole),
        ureg.joule / (ureg.kelvin * ureg.mole)
    ) == (R_, ureg.joule / (ureg.kelvin * ureg.mole))

    # TEST CASE 5: Strip unit from Quantity with non-standard
    # incompatible unit specified
    with pytest.raises(ValueError):
        strip_unit(
            R_ * unit.joule / (unit.kelvin * unit.mole),
            ureg.meter ** 3 * ureg.atmosphere / (ureg.kelvin * ureg.mole)
        )


"""
OpenMM physical constants and unit conversions
==============================================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains physical constants and functions for unit
reduction.
"""

from openmm import unit

from ..algorithm import unit as u

VACUUM_PERMITTIVITY = 8.854187812813e-12 * unit.farad / unit.meter


def get_scale_factors(
    bases: dict[str, unit.Quantity], other: dict[str, list] = {}
) -> dict[str, unit.Quantity]:
    """
    Evaluates scaling factors for reduced units.

    .. seealso::

       This function is an alias for
       :func:`mdcraft.algorithm.unit.get_scale_factors`.
    """

    return u.get_scale_factors(bases, other)


def get_lj_scale_factors(
    bases: dict[str, unit.Quantity], other: dict[str, list] = {}
) -> dict[str, unit.Quantity]:
    """
    Evaluates scaling factors for reduced Lennard-Jones units.

    .. seealso::

       This function is an alias for
       :func:`mdcraft.algorithm.unit.get_lj_scale_factors`.
    """

    return u.get_lj_scale_factors(bases, other)



"""
Accelerated algorithms
======================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains miscellaneous Numba-accelerated common algorithms.
"""

import numba
import numpy as np


@numba.njit(fastmath=True)
def numba_histogram_bin_edges(
    array: np.ndarray[float], n_bins: int
) -> np.ndarray[float]:
    r"""
    Serial Numba-accelerated function to compute the uniform histogram
    bin edges for a one-dimensional NumPy array :math:`\mathbf{a}`
    and a specified number of bins :math:`N_\mathrm{bins}`.

    Parameters
    ----------
    array : `np.ndarray`
        One-dimensional array :math:`\mathbf{a}`.

    n_bins : `int`
        Number of bins :math:`N_\mathrm{bins}`.

    Returns
    -------
    bin_edges : `np.ndarray`
        Uniform histogram bin edges for the array :math:`\mathbf{a}`.
    """

    min_, max_ = array.min(), array.max()
    n_edges = n_bins + 1
    bin_edges = np.empty(n_edges)
    delta = (max_ - min_) / n_bins
    for i in range(n_edges):
        bin_edges[i] = min_ + i * delta
    bin_edges[-1] = max_
    return bin_edges


@numba.njit(fastmath=True)
def numba_histogram(
    array: np.ndarray[float], n_bins: int, bin_edges: np.ndarray[float]
) -> np.ndarray[int]:
    r"""
    Serial Numba-accelerated function to compute the histogram of a
    one-dimensional NumPy array :math:`\mathbf{a}` using predetermined
    bin edges for :math:`N_\mathrm{bins}` bins.

    Parameters
    ----------
    array : `np.ndarray`
        One-dimensional array :math:`\mathbf{a}`.

    n_bins : `int`
        Number of bins :math:`N_\mathrm{bins}`.

    bin_edges : `np.ndarray`
        Bin edges.

    Returns
    -------
    histogram_ : `np.ndarray`
        Histogram of the array :math:`\mathbf{a}`.
    """

    min_, max_ = bin_edges[0], bin_edges[-1]
    histogram_ = np.zeros(n_bins, dtype=np.intp)
    for x in array:
        if x == max_:
            bin_ = n_bins - 1
        else:
            bin_ = int(n_bins * (x - min_) / (max_ - min_))
        if 0 <= bin_ < n_bins:
            histogram_[bin_] += 1
    return histogram_


@numba.njit(fastmath=True)
def numba_dot(a: np.ndarray[float], b: np.ndarray[float]) -> float:
    r"""
    Serial Numba-accelerated dot product between two one-dimensional
    NumPy arrays :math:`\mathbf{a}` and :math:`\mathbf{b}`, each with
    shape :math:`(3,)`.

    .. math::

       \mathbf{a}\cdot\mathbf{b}=a_1b_1+a_2b_2+a_3b_3

    Parameters
    ----------
    a : `np.ndarray`
        First vector :math:`\mathbf{a}`.

        **Shape**: :math:`(3,)`.

    b : `np.ndarray`
        Second vector :math:`\mathbf{b}`.

        **Shape**: :math:`(3,)`.

    Returns
    -------
    ab : `float`
        Dot product :math:`\mathbf{a}\cdot\mathbf{b}`.
    """

    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


@numba.njit(fastmath=True)
def numba_inner(qs: np.ndarray[float], rs: np.ndarray[float]) -> np.ndarray[float]:
    r"""
    Serial Numba-accelerated inner product between all possible
    combinations of multiple one-dimensional NumPy arrays
    :math:`\mathbf{q}` and :math:`\mathbf{r}`, each with shape
    :math:`(3,)`.

    .. math::

       \mathbf{q}_i\cdot\mathbf{r}_j
       =q_{i1}r_{j1}+q_{i2}r_{j2}+q_{i3}r_{j3}

    Parameters
    ----------
    qs : `np.ndarray`
        Multiple vectors :math:`\mathbf{q}`.

        **Shape**: :math:`(N_q,\,3)`.

    rs : `np.ndarray`
        Multiple vectors :math:`\mathbf{r}`.

        **Shape**: :math:`(N_r,\,3)`.

    Returns
    -------
    s : `np.ndarray`
        Inner products of the vectors,
        :math:`\mathbf{q}_i\cdot\mathbf{r}_j`.

        **Shape**: :math:`(N_q,\,N_r)`.
    """

    s = np.empty((qs.shape[0], rs.shape[0]))
    for i in range(qs.shape[0]):
        for j in range(rs.shape[0]):
            s[i, j] = numba_dot(qs[i], rs[j])
    return s


@numba.njit(fastmath=True, parallel=True)
def numba_inner_parallel(
    qs: np.ndarray[float], rs: np.ndarray[float]
) -> np.ndarray[float]:
    r"""
    Parallel Numba-accelerated inner product between all possible
    combinations of multiple one-dimensional NumPy arrays
    :math:`\mathbf{q}` and :math:`\mathbf{r}`, each with shape
    :math:`(3,)`.

    .. math::

       \mathbf{q}_i\cdot\mathbf{r}_j
       =q_{i1}r_{j1}+q_{i2}r_{j2}+q_{i3}r_{j3}

    Parameters
    ----------
    qs : `np.ndarray`
        Multiple vectors :math:`\mathbf{q}`.

        **Shape**: :math:`(N_q,\,3)`.

    rs : `np.ndarray`
        Multiple vectors :math:`\mathbf{r}`.

        **Shape**: :math:`(N_r,\,3)`.

    Returns
    -------
    s : `np.ndarray`
        Inner products of the vectors,
        :math:`\mathbf{q}_i\cdot\mathbf{r}_j`.

        **Shape**: :math:`(N_q,\,N_r)`.
    """

    s = np.empty((qs.shape[0], rs.shape[0]))
    for i in numba.prange(qs.shape[0]):
        for j in range(rs.shape[0]):
            s[i, j] = numba_dot(qs[i], rs[j])
    return s



import pathlib
import sys

import MDAnalysis as mda
from MDAnalysis.tests.datafiles import DCD, PSF
import numpy as np
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.algorithm import molecule # noqa: E402

# Load sample topology and trajectory
universe = mda.Universe(PSF, DCD)
protein = universe.select_atoms("protein")
core = universe.select_atoms("protein and (resid 1-29 60-121 160-214)")
nmp = universe.select_atoms("protein and resid 30-59")
lid = universe.select_atoms("protein and resid 122-159")
arg = universe.select_atoms("resname ARG")

def test_func_center_of_mass():

    """
    Test cases 5–15 are inspired by the example in the "Working with
    AtomGroups" section of the MDAnalysis Tutorial
    (https://www.mdanalysis.org/MDAnalysisTutorial/atomgroups.html).
    """

    # TEST CASE 1: No topology or trajectory
    with pytest.raises(ValueError):
        molecule.center_of_mass()

    # TEST CASE 2: Invalid grouping
    with pytest.raises(ValueError):
        molecule.center_of_mass(universe.atoms, "atoms")

    # TEST CASE 3: No system dimension information when number of periodic
    # boundary crossings is provided
    with pytest.raises(ValueError):
        molecule.center_of_mass(universe.atoms,
                                images=np.zeros((universe.atoms.n_atoms, 3)))

    # TEST CASE 4: Incompatible mass and position arrays
    with pytest.raises(ValueError):
        molecule.center_of_mass(
            masses=universe.atoms.masses,
            positions=[r.atoms.positions for r in universe.residues]
        )

    # TEST CASE 5: Center of mass of domains in AdK
    assert np.allclose(molecule.center_of_mass(core), core.center_of_mass())
    assert np.allclose(molecule.center_of_mass(nmp), nmp.center_of_mass())
    assert np.allclose(molecule.center_of_mass(lid), lid.center_of_mass())

    # TEST CASE 6: Center of mass of all particles using AtomGroup
    com = universe.atoms.center_of_mass()
    assert np.allclose(molecule.center_of_mass(universe.atoms), com)

    # TEST CASE 7: Center of mass of all particles using AtomGroup, but
    # with the masses and unwrapped positions returned
    c, m, p = molecule.center_of_mass(
        universe.atoms,
        images=np.zeros((universe.atoms.n_atoms, 3), dtype=int),
        dimensions=np.array((0, 0, 0)),
        raw=True
    )
    assert np.allclose(c, com)
    assert np.allclose(m, universe.atoms.masses)
    assert np.allclose(p, universe.atoms.positions)

    # TEST CASE 8: Centers of mass of different residues using AtomGroup
    res_coms = np.array([r.atoms.center_of_mass()
                         for r in universe.residues])
    assert np.allclose(molecule.center_of_mass(universe.atoms, "residues"),
                       res_coms)

    # TEST CASE 9: Centers of mass of different residues using raw masses
    # and positions from AtomGroup
    assert np.allclose(
        molecule.center_of_mass(
            universe.atoms, "residues",
            masses=[r.atoms.masses for r in universe.residues]
        ),
        res_coms
    )

    # TEST CASE 10: Centers of mass of different residues using raw masses
    # and positions
    assert np.allclose(
        molecule.center_of_mass(
            masses=[r.atoms.masses for r in universe.residues],
            positions=[r.atoms.positions for r in universe.residues]
        ),
        res_coms
    )

    # TEST CASE 11: Centers of mass of arginine residues using AtomGroup
    arg_coms = np.array([r.atoms.center_of_mass() for r in arg.residues])
    assert np.allclose(molecule.center_of_mass(arg, "residues"), arg_coms)

    # TEST CASE 12: Centers of mass of arginine residues using AtomGroup
    # and specified number of residues
    assert np.allclose(molecule.center_of_mass(arg, n_groups=13), arg_coms)

    # TEST CASE 13: Centers of mass of arginine residues using raw masses
    # and positions
    assert np.allclose(molecule.center_of_mass(masses=arg.masses,
                                               positions=arg.positions,
                                               n_groups=13),
                       arg_coms)

    # TEST CASE 14: Centers of mass of only segment in AtomGroup
    assert np.allclose(molecule.center_of_mass(universe.atoms, "segments"),
                       com)

    # TEST CASE 15: Centers of mass of only segment in AtomGroup
    # containing the arginine residues
    assert np.allclose(molecule.center_of_mass(arg, "segments"),
                       arg.center_of_mass())

def test_radius_of_gyration():

    """
    The reference implementation is adapted from the "Writing your own
    trajectory analysis" section of the MDAnalysis User Guide
    (https://userguide.mdanalysis.org/stable/examples/analysis/custom_trajectory_analysis.html).
    """

    def radius_of_gyration(group):
        positions = group.positions
        masses = group.masses
        center_of_mass = group.center_of_mass()
        r_sq = (positions - center_of_mass) ** 2
        r_ssq = np.array((r_sq.sum(axis=1),
                          (r_sq[:, [1, 2]]).sum(axis=1),
                          (r_sq[:, [0, 2]]).sum(axis=1),
                          (r_sq[:, [0, 1]]).sum(axis=1)))
        return np.sqrt((masses * r_ssq).sum(axis=1) / masses.sum())

    # TEST CASE 1: Invalid grouping
    with pytest.raises(ValueError):
        molecule.radius_of_gyration(universe.atoms, "atoms")

    # TEST CASE 2: Overall radius of gyration
    ref = radius_of_gyration(universe.atoms)
    assert np.isclose(molecule.radius_of_gyration(universe.atoms), ref[0])
    assert np.allclose(molecule.radius_of_gyration(universe.atoms,
                                                   components=True),
                       ref[1:])

    # TEST CASE 3: Radii of gyration of arginine residues
    ref = np.array([radius_of_gyration(g.atoms) for g in arg.residues])
    assert np.allclose(molecule.radius_of_gyration(arg, "residues"),
                       ref[:, 0])
    assert np.allclose(molecule.radius_of_gyration(arg, "residues",
                                                   components=True),
                       ref[:, 1:])

    # TEST CASE 4: Radii of gyration of different residues
    ref = np.array([radius_of_gyration(g.atoms) for g in universe.residues])
    assert np.allclose(molecule.radius_of_gyration(universe.atoms, "residues"),
                       ref[:, 0])
    assert np.allclose(molecule.radius_of_gyration(universe.atoms, "residues",
                                                   components=True),
                       ref[:, 1:])


import pathlib
import sys

from openmm import unit

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.openmm import unit as u # noqa: E402

def test_func_lj_scaling():

    temp = 300 * unit.kelvin

    # TEST CASE 1: Correct units for complex scaling factors
    scales = u.get_lj_scale_factors(
        {"mass": 18.0153 * unit.amu,
         "length": 0.275 * unit.nanometer,
         "energy": (unit.BOLTZMANN_CONSTANT_kB * temp).in_units_of(unit.kilojoule)}
    )
    assert scales["molar_energy"].unit == unit.kilojoule_per_mole
    assert scales["velocity"].unit == unit.nanometer / unit.picosecond
    assert scales["electric_field"].unit \
           == unit.kilojoule_per_mole / (unit.nanometer * unit.elementary_charge)

    # TEST CASE 2: No default scaling factors
    scales = u.get_scale_factors(
        {"mass": 18.0153 * unit.amu,
         "length": 0.275 * unit.nanometer,
         "energy": (unit.BOLTZMANN_CONSTANT_kB * temp).in_units_of(unit.kilojoule),
         "charge": 1 * unit.elementary_charge},
        {"surface_charge_density": (("charge", 1), ("length", -2))}
    )
    assert "time" not in scales

    # TEST CASE 3: Custom scaling factors
    assert scales["surface_charge_density"].unit == unit.elementary_charge / unit.nanometer ** 2


"""
OpenMM utility functions
========================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains OpenMM-related utility functions.
"""

from datetime import datetime
import itertools
import logging
from typing import Union

import numpy as np
import openmm
from openmm import unit


def _create_context(
    system: openmm.System,
    integrator: openmm.Integrator,
    positions: np.ndarray[float],
    platform: openmm.Platform,
    properties: dict,
) -> openmm.Context:
    r"""
    Creates an OpenMM Context by cloning the Integrator passed to this
    function. Useful for benchmarking different simulation systems.

    Parameters
    ----------
    system : `openmm.System`
        OpenMM molecular system.

    integrator : `openmm.Integrator`
        OpenMM integrator or thermostat.

    positions : `np.ndarray` or `unit.Quantity`
        Initial positions of the :math:`N` particles in the system.

        **Shape**: :math:`(N,\,3)`.

        **Reference unit**: :math:`\mathrm{nm}`.

    platform : `openmm.Platform`
        OpenMM platform.

    properties : `dict`
        Dictionary of platform-specific properties.

    Returns
    -------
    context : `openmm.Context`
        OpenMM simulation context.
    """

    integrator = openmm.XmlSerializer.clone(integrator)
    context = openmm.Context(system, integrator, platform, properties)
    context.setPositions(positions)
    return context


def _benchmark_integrator(context: openmm.Context, steps: int) -> float:
    """
    Benchmarks the performance of an OpenMM Integrator.

    Parameters
    ----------
    context : `openmm.Context`
        OpenMM simulation context.

    Returns
    -------
    time : `float`
        Elapsed time, in seconds.
    """

    start = datetime.now()
    context.getIntegrator().step(steps)
    return (datetime.now() - start).total_seconds()


def optimize_pme(
    system: openmm.System,
    integrator: openmm.Integrator,
    positions: Union[np.ndarray[float], unit.Quantity],
    platform: openmm.Platform,
    properties: dict,
    min_cutoff: Union[float, unit.Quantity],
    max_cutoff: Union[float, unit.Quantity],
    *,
    pmeforce: openmm.NonbondedForce = None,
    cpu_pme: bool = True,
    target: float = 10,
    target_std: float = None,
    window: int = 3,
    fastest: int = 5,
    rerun: int = 2,
    verbose: bool = True,
) -> tuple[unit.Quantity, bool]:
    """
    Runs a series of simulations using different parameters to determine
    the optimal configuration for evaluating electrostatic interactions
    with the particle mesh Ewald (PME) method on a GPU (CUDA or OpenCL).

    The cutoff distance for the Coulomb potential can be freely varied
    with no accuracy penalty since OpenMM automatically selects internal
    parameters to satisfy the specified error tolerance. However, it may
    affect the accuracy of other nonbonded interactions, so care must be
    taken to ensure the optimal Coulomb potential cutoff is compatible
    with the other pair potentials in the system.

    In certain cases, OpenMM can perform better with the reciprocal
    space calculations being done on the CPU while the direct space
    calculations are being evaluated on the GPU.

    Parameters
    ----------
    system : `openmm.System`
        OpenMM molecular system.

    integrator : `openmm.Integrator`
        OpenMM integrator or thermostat.

    positions : `np.ndarray` or `unit.Quantity`
        Initial positions of the :math:`N` particles in the system.

        **Shape**: :math:`(N,\\,3)`.

        **Reference unit**: :math:`\\mathrm{nm}`.

    platform : `openmm.Platform`
        OpenMM platform.

    properties : `dict`
        Dictionary of platform-specific properties.

    min_cutoff : `float` or `unit.Quantity`
        Minimum cutoff distance to test.

        **Reference unit**: :math:`\\mathrm{nm}`.

    max_cutoff : `float` or `unit.Quantity`
        Maximum cutoff distance to test.

        **Reference unit**: :math:`\\mathrm{nm}`.

    pmeforce : `openmm.NonbondedForce` or `openmm.AmoebaMultipoleForce`, \
    keyword-only, optional
        Pair potential with electrostatic interactions that are
        evaluated using PME.

    cpu_pme : `bool`, keyword-only, default: :code:`True`
        Determines whether CPU PME should be benchmarked.

    target : `float`, keyword-only, default: :code:`10`
        Target simulation time for each test run, in seconds.

    target_std : `float`, keyword-only, optional
        Allowed variability for `target`, in seconds. If set to a value
        that is too small, the target simulation time may never be
        satisfied. If not specified, it is set to 10% of `target`.

    window : `int`, keyword-only, default: :code:`3`
        Number of previous runs to look at before deciding whether to
        stop testing larger cutoffs.

    fastest : `int`, keyword-only, default: :code:`5`
        Number of fastest configurations to retest before deciding on
        the optimal cutoff distance and hardware (architecture).

    rerun : `int`, keyword-only, default :code:`2`
        Number of reruns to complete for the fastest preliminary runs.

    verbose : `bool`, keyword-only, default: :code:`True`
        Determines whether detailed progress is shown.

    Returns
    -------
    cutoff : `unit.Quantity`
        Optimal cutoff distance.

        **Reference unit**: :math:`\\mathrm{nm}`.

    cpu_pme : `bool`
        Specifies whether to use the CPU to perform reciprocal space
        calculations.
    """

    # Set up logger
    logging.basicConfig(
        format="{asctime} | {levelname:^8s} | {message}",
        style="{",
        level=logging.INFO if verbose else logging.WARNING,
    )

    # Get information about the pair potential of interest
    if pmeforce is None:
        for force in system.getForces():
            if isinstance(force, (openmm.NonbondedForce, openmm.AmoebaMultipoleForce)):
                pmeforce = force
                break
    if pmeforce.getNonbondedMethod() != openmm.NonbondedForce.PME:
        raise ValueError(
            "The provided (or guessed) pair potential is "
            "not being evaluated using the particle mesh "
            "Ewald (PME) method."
        )
    cpu_pme &= isinstance(pmeforce, openmm.NonbondedForce) and platform.supportsKernels(
        ["CalcPmeReciprocalForce"]
    )
    tol = pmeforce.getEwaldErrorTolerance()

    # Determine the number of timesteps to run for each cutoff distance
    logging.info("Determining a reasonable number of timesteps for PME optimizer...")
    pmeforce.setCutoffDistance(np.sqrt(min_cutoff * max_cutoff))
    properties["UseCpuPme"] = "false"
    context = _create_context(system, integrator, positions, platform, properties)
    if target_std is None:
        target_std = 0.1 * target
    time_width = max(9, np.ceil(np.log10(target)).astype(int) + 7)
    lb, ub = target - target_std, target + target_std
    steps, time = 20, 0
    while True:
        time = _benchmark_integrator(context, steps)
        logging.info(f"  GPU: {steps:14,} ts " f"===> {time:{time_width}.5f} s elapsed")
        if lb < time < ub:
            break
        steps = int(target * steps / time)
    if cpu_pme:
        properties["UseCpuPme"] = "true"
        context = _create_context(system, integrator, positions, platform, properties)
        steps_cpu, time = 20, 0
        while True:
            time = _benchmark_integrator(context, steps_cpu)
            logging.info(
                f"  CPU: {steps_cpu:14,} ts " f"===> {time:{time_width}.5f} s elapsed"
            )
            if lb < time < ub:
                break
            steps_cpu = int(target * steps_cpu / time)
        steps = min(steps, steps_cpu)
    steps = np.round(steps, 2 - np.ceil(np.log10(steps)).astype(int))
    logging.info(f"Starting PME optimizer (using {steps:,} timesteps)...")

    # Build list of cutoff distances to test
    if isinstance(min_cutoff, unit.Quantity):
        min_cutoff = min_cutoff.value_in_unit(unit.nanometer)
    if isinstance(max_cutoff, unit.Quantity):
        max_cutoff = max_cutoff.value_in_unit(unit.nanometer)
    cutoffs = {"gpu": {min_cutoff}}
    if cpu_pme:
        cutoffs["cpu"] = {min_cutoff}

    for dim in [
        v[i].value_in_unit(unit.nanometer)
        for i, v in enumerate(system.getDefaultPeriodicBoxVectors())
    ]:

        # Iterate through possible grid sizes
        for n_mesh in itertools.count(start=5):

            # Check if grid size is legal for FFT
            check = n_mesh
            for factor in (2, 3, 5, 7):
                while check > 1 and check % factor == 0:
                    check /= factor
            if check not in (1, 11, 13):
                continue

            # Compute smallest cutoff that will give the current grid
            # size
            alpha = 1.5 * n_mesh * tol**0.2 / dim
            cutoff = np.round(np.sqrt(-np.log(2 * tol) / alpha), 3)
            if cutoff < min_cutoff:
                break
            if cutoff < max_cutoff:
                if cpu_pme:
                    cutoffs["cpu"].add(cutoff)
                if check == 1:
                    cutoffs["gpu"].add(cutoff)

    # Get preliminary times for the different architectures and cutoffs
    cutoff_width = max(
        7, np.ceil(np.log10(max(max(v) for v in cutoffs.values()))).astype(int) + 6
    )
    times = {}
    for arch, cut in cutoffs.items():
        cutoffs[arch] = np.array(sorted(cut))
        times[arch] = np.empty(cutoffs[arch].shape)
        times[arch][:] = np.nan
        for i, cutoff in enumerate(cutoffs[arch]):
            pmeforce.setCutoffDistance(cutoff)
            properties["UseCpuPme"] = str(arch == "cpu").lower()
            context = _create_context(
                system, integrator, positions, platform, properties
            )
            times[arch][i] = _benchmark_integrator(context, steps)
            logging.info(
                f"  {arch.upper()}: {cutoff:{cutoff_width}.4f} nm cutoff "
                f"===> {times[arch][i]:{time_width}.5f} s elapsed"
            )

            # Stop iteration if simulation is continuously getting slower
            if i > window and np.all(
                times[arch][i - window : i] > times[arch][i - window - 1 : i - 1]
            ):
                break
    best = sorted(
        [t, c, a] for a in times.keys() for c, t in zip(cutoffs[a], times[a])
    )[:fastest]

    # Rerun the fastest configurations to ensure correct results
    for i, (time, cutoff, arch) in enumerate(best):
        pmeforce.setCutoffDistance(cutoff)
        properties["UseCpuPme"] = str(arch == "cpu").lower()
        context = _create_context(system, integrator, positions, platform, properties)

        # Replace preliminary time with median time from reruns
        best[i][0] = sorted(
            (time, *[_benchmark_integrator(context, steps) for _ in range(rerun)])
        )[1]
    best.sort()

    # Display fastest configurations and timings
    time_width = 8 + 2 * np.ceil(max(0, time_width - 8) // 2).astype(int)
    cutoff_width = 11 + 2 * np.ceil(max(0, cutoff_width - 11) // 2).astype(int)
    logging.info(
        f"PME optimization completed.\n"
        f"   Rank | {'Time (s)':^{time_width}} "
        f"| {'Cutoff (nm)':^{cutoff_width}} | CPU PME\n"
        f"  ------|{'-' * (time_width + 2)}"
        f"|{'-' * (cutoff_width + 2)}|---------\n  "
        + "\n  ".join(
            f" {i + 1:>4} | {time:{time_width}.5f} |"
            f" {cutoff:{cutoff_width}.4f} | {arch == 'cpu'}"
            for i, (time, cutoff, arch) in enumerate(best)
        )
    )

    # Return optimized configuration
    return best[0][1] * unit.nanometer, arch == "cpu"



"""
Utility algorithms
==================
.. moduleauthor:: Benjamin Ye <GitHub: @bbye98>

This module contains logical and mathematical utility functions used by
other MDCraft modules.
"""

from typing import Any

import numpy as np
import sympy


def get_closest_factors(
    value: int, n_factors: int, reverse: bool = False
) -> np.ndarray[int]:
    """
    Finds the :math:`n` closest factors for a given number :math:`N`,
    sorted in ascending order.

    Parameters
    ----------
    value : `int`
        Number :math:`N` to be factorized.

    n_factors : `int`
        Number of factors :math:`n` to return.

    reverse : `bool`, optional, default: :code:`False`
        Specifies whether to sort in descending order.

    Returns
    -------
    factors : `np.ndarray`
        :math:`n` closest factors for `N`.

        **Shape**: :math:`(n,)`.
    """

    # Take the n-th root of N
    rt = value ** (1 / n_factors)
    rt_int = int(np.round(rt))
    if np.isclose(rt, rt_int):
        return rt_int * np.ones(n_factors, dtype=int)

    # Get all factors of N
    _factors = np.fromiter(
        (
            factor
            for factor, power in sympy.ntheory.factorint(value).items()
            for _ in range(power)
        ),
        dtype=int,
    )

    # Find n closest factors
    i = 0
    factors = np.ones(n_factors, dtype=int)
    for j, f in enumerate(_factors[::-1]):
        while True:
            if i < n_factors:
                m = factors[i] * f
                if m <= rt_int or j < n_factors and factors[i] == 1:
                    factors[i] = m
                    break
                i += 1
            else:
                factors[np.argmin(factors)] *= f
                break

    if reverse:
        return np.sort(factors)[::-1]
    return np.sort(factors)


def replicate(
    cell_dims: np.ndarray[float], cell_pos: np.ndarray[float], n_cells: np.ndarray[int]
) -> np.ndarray[float]:
    r"""
    Replicates points in an unit cell along the :math:`x`-, :math:`y`-,
    and :math:`z`-directions.

    Parameters
    ----------
    cell_dims : `numpy.ndarray`
        Dimensions of the unit cell.

        **Shape**: :math:`(3,)`.

    cell_pos : `numpy.ndarray`
        Positions of the :math:`N` points inside the unit cell.

        **Shape**: :math:`(N,\,3)`.

    n_cells : `numpy.ndarray`
        Number of times to replicate the unit cell in each direction.

        **Shape**: :math:`(3,)`.

    Returns
    -------
    pos : `numpy.ndarray`
        Positions of the original and replicated points.
    """

    # Add cell x-dimensions to cell x-positions and replicate them
    # n_y * n_z times
    x = np.tile(
        np.concatenate(
            cell_pos[:, 0] + (cell_dims[0] * np.arange(n_cells[0]))[:, None]
        ),
        reps=n_cells[1] * n_cells[2],
    )

    # Replicate cell y-positions n_x times, add cell y-dimensions to
    # them, and then replicate them n_z times
    y = np.tile(
        np.concatenate(
            np.tile(cell_pos[:, 1], reps=n_cells[0])
            + (np.arange(n_cells[1]) * cell_dims[1])[:, None]
        ),
        reps=n_cells[2],
    )

    # Replicate cell z-positions n_x * n_y times and add cell
    # z-dimensions to them
    z = np.concatenate(
        np.tile(cell_pos[:, 2], reps=n_cells[0] * n_cells[1])
        + cell_dims[2] * np.arange(n_cells[2])[:, None]
    )

    return np.vstack((x, y, z)).T


def rebin(x: np.ndarray[float], factor: int = None) -> np.ndarray[float]:
    r"""
    Rebins discrete data.

    Parameters
    ----------
    x : `numpy.ndarray`
        Discrete data to be rebinned in the last dimension.

    factor : `int`, optional
        Size reduction factor. If not specified, the biggest factor
        on the order of :math:`\mathcal{O}(1)`, if available, is used.

    Returns
    -------
    xr : `numpy.ndarray`
        Rebinned discrete data.
    """

    if factor is None:
        factors = np.array(sympy.divisors(x.shape[-1])[1:])
        factor_indices = np.where(factors < 10)[0]
        if len(factor_indices):
            factor = factors[factor_indices[-1]]
        else:
            raise ValueError("No factor provided for rebinning.")

    return x.reshape((*x.shape[:-1], -1, factor)).mean(axis=-1)


def depth_first_search(
    graph: dict[Any, list[Any]], start: Any, visited: bool, group: list[Any]
) -> None:
    """
    Implements the depth-first search algorithm to find connected
    components in a graph.

    Parameters
    ----------
    graph : `dict`
        Graph to search.

    start : `Any`
        Starting node.

    visited : `dict`
        Visited nodes. This `dict` is updated in-place.

    group : `list`
        Connected nodes. This `list` is updated in-place.
    """

    visited[start] = True
    group.append(start)
    for neighbor in graph[start]:
        if not visited[neighbor]:
            depth_first_search(graph, neighbor, visited, group)


def find_connected_nodes(graph: dict[Any, list[Any]]) -> list[list[Any]]:
    """
    Finds connected components in a graph.

    Parameters
    ----------
    graph : `dict`
        Graph to search.

    Returns
    -------
    results : `list`
        Connected components.
    """

    visited = {node: False for node in graph}
    results = []
    for start in graph:
        if not visited[start]:
            group = []
            depth_first_search(graph, start, visited, group)
            results.append(group)
    return results


def is_lower_triangular(matrix: np.ndarray[float]) -> bool:
    """
    Checks if a matrix is lower triangular.

    Parameters
    ----------
    matrix : `numpy.ndarray`
        Matrix to check.

    Returns
    -------
    is_lower_triangular : `bool`
        Whether the matrix is lower triangular.
    """

    return np.allclose(matrix, np.tril(matrix))



import pathlib
import sys

import numpy as np
from openmm import app, unit
import pytest

sys.path.insert(0, f"{pathlib.Path(__file__).parents[1].resolve().as_posix()}/src")
from mdcraft.algorithm import topology # noqa: E402

rng = np.random.default_rng()
dims = np.array((10.0, 10.0, 10.0))

def test_func_create_atoms():

    # TEST CASE 1: N not specified
    with pytest.raises(ValueError):
        topology.create_atoms(dims)

    # TEST CASE 2: N not an integer
    with pytest.raises(ValueError):
        topology.create_atoms(dims, np.pi)

    # TEST CASE 3: Invalid N_p
    with pytest.raises(ValueError):
        topology.create_atoms(dims, 9000, 9001)

    # TEST CASE 4: N not divisible by N_p
    with pytest.raises(ValueError):
        topology.create_atoms(dims, 10, 3)

    # TEST CASE 5: Random melt in reduced units
    N = rng.integers(1, 1000)
    pos = topology.create_atoms(dims, N)
    assert pos.shape == (N, 3)

    # TEST CASE 6: Random melt with default length unit
    pos = topology.create_atoms(dims * unit.nanometer, N)
    assert pos.shape == (N, 3) and pos.unit == unit.nanometer

    # TEST CASE 7: Random melt with specific length unit
    pos = topology.create_atoms(dims * unit.nanometer, N, length_unit=unit.angstrom)
    assert pos.shape == (N, 3) and pos.unit == unit.angstrom

    # TEST CASE 8: Topology provided instead of dimensions
    topo = app.Topology()
    topo.setUnitCellDimensions(dims)
    pos = topology.create_atoms(topo, N)
    assert pos.shape == (N, 3)

    # TEST CASE 9: Random polymer melt
    M = rng.integers(1, 100)
    N_p = rng.integers(4, 100)
    N = M * N_p
    pos = topology.create_atoms(dims, N, N_p)
    assert pos.shape == (N, 3)

    # TEST CASE 10: Random polymer melt with bond, angle, and dihedral
    # information and wrapped positions
    pos, bonds, angles, dihedrals = topology.create_atoms(
        dims, N, N_p, bonds=True, angles=True, dihedrals=True, randomize=True,
        wrap=True
    )
    assert pos.shape == (N, 3)
    assert bonds.shape[0] == N - M
    assert angles.shape[0] == N - 2 * M
    assert dihedrals.shape[0] == N - 3 * M
    assert np.all((pos[:, 0] > 0) & (pos[:, 0] < dims[0]))
    assert np.all((pos[:, 1] > 0) & (pos[:, 2] < dims[1]))
    assert np.all((pos[:, 1] > 0) & (pos[:, 2] < dims[2]))

    # TEST CASE 11: FCC lattice with flexible dimensions
    pos, new_dims = topology.create_atoms(dims, lattice="fcc", length=0.8,
                                          flexible=True)
    assert np.allclose(pos[4], 0.8 * np.array((0, np.sqrt(3) / 3, 2 * np.sqrt(6) / 3)))
    assert np.allclose(dims, new_dims, atol=1)

    # TEST CASE 12: HCP lattice with flexible dimensions
    pos, new_dims = topology.create_atoms(dims, lattice="hcp", length=0.8,
                                          flexible=True)
    assert np.allclose(pos[1], 0.8 * np.array((0.5, np.sqrt(3) / 2, 0)))
    assert np.allclose(dims, new_dims, atol=1)

    # TEST CASE 13: HCP lattice to fill specified dimensions
    pos, new_dims = topology.create_atoms(dims, lattice="hcp", length=0.8)
    assert np.allclose(pos[1], 0.8 * np.array((0.5, np.sqrt(3) / 2, 0)))
    assert np.allclose(dims, new_dims, atol=1)

    # TEST CASE 14: Graphene wall
    pos, new_dims = topology.create_atoms(dims, lattice="honeycomb",
                                          length=0.142 * unit.nanometer,
                                          flexible=True)
    assert pos[1, 1] == 0.142 * unit.nanometer
    assert np.allclose(dims[:2], new_dims[:2], atol=1)
    assert new_dims[2] == 0 * unit.nanometer

    # TEST CASE 15: Cubic crystal lattice
    pos, new_dims = topology.create_atoms(dims, lattice="cubic", length=1)
    assert np.allclose(pos[-1], dims - 1)

    # TEST CASE 16: HCP wall
    pos, new_dims = topology.create_atoms([10, 10, 0], lattice="hcp", length=1, flexible=True)
    assert np.allclose(pos[:, 2], 0) and new_dims[2] == 0

def test_func_unwrap():

    pos_old = np.array(((2.0, 2.0, 2.0),))
    images = np.zeros_like(pos_old, dtype=int)
    thresholds = dims / 2
    pos = np.array(((8.0, 8.0, 8.0),))

    # TEST CASE 1: Unwrap not in-place
    pos_unwrapped, pos_old_updated, images = topology.unwrap(
        pos, pos_old, dims, thresholds=thresholds, images=images,
        in_place=False
    )
    assert (np.allclose(pos_unwrapped[0], -2)
            and np.allclose(pos, pos_old_updated))

    # TEST CASE 2: Unwrap in-place
    topology.unwrap(pos, pos_old, dims)
    assert np.allclose(pos[0], -2)

def test_func_wrap():

    pos = np.array(((9.0, 10.0, 11.0),))

    # TEST CASE 1: Wrap not in-place
    pos_wrapped = topology.wrap(pos, dims, in_place=False)
    assert np.allclose(pos_wrapped[0], (9, 10, 1))

    # TEST CASE 2: Wrap in-place
    topology.wrap(pos, dims)
    assert np.allclose(pos[0], (9, 10, 1))

def test_func_convert_cell_representation():

    # TEST CASE 1: Cubic box
    parameters = np.array((1, 1, 1, 90, 90, 90))
    vectors = np.eye(3)
    assert np.allclose(
        topology.convert_cell_representation(vectors=vectors),
        parameters
    )
    assert np.allclose(
        topology.convert_cell_representation(parameters=parameters),
        vectors
    )

    # TEST CASE 2: Rhombic docecahedron (xy-square)
    parameters = np.array((1, 1, 1, 60, 60, 90))
    vectors = np.array((
        (1, 0, 0),
        (0, 1, 0),
        (1 / 2, 1 / 2, 1 / np.sqrt(2))
    ))
    assert np.allclose(
        topology.convert_cell_representation(vectors=vectors),
        parameters
    )
    assert np.allclose(
        topology.convert_cell_representation(parameters=parameters),
        vectors
    )

    # TEST CASE 3: Rhombic docecahedron (xy-hexagon)
    parameters = np.array((1, 1, 1, 60, 60, 60))
    vectors = np.array((
        (1, 0, 0),
        (1 / 2, np.sqrt(3) / 2, 0),
        (1 / 2, np.sqrt(3) / 6, np.sqrt(6) / 3)
    ))
    assert np.allclose(
        topology.convert_cell_representation(vectors=vectors),
        parameters
    )
    assert np.allclose(
        topology.convert_cell_representation(parameters=parameters),
        vectors
    )

    # TEST CASE 4: Truncated octahedron
    parameters = np.array((1, 1, 1, 
                           np.degrees(np.arccos(1 / 3)), 
                           np.degrees(np.arccos(-1 / 3)), 
                           np.degrees(np.arccos(1 / 3))))
    vectors = np.array((
        (1, 0, 0),
        (1 / 3, 2 * np.sqrt(2) / 3, 0),
        (-1 / 3, np.sqrt(2) / 3, np.sqrt(6) / 3)
    ))
    assert np.allclose(
        topology.convert_cell_representation(vectors=vectors),
        parameters
    )
    assert np.allclose(
        topology.convert_cell_representation(parameters=parameters),
        vectors
    )